from __future__ import annotations

from functools import reduce
from logging import getLogger
from re import (
    DOTALL,
    IGNORECASE,
    MULTILINE,
    Match as RegexMatch,
    Pattern,
    compile as regexp,
    error as RegexError,  # noqa: N812
)
from textwrap import dedent
from typing import TYPE_CHECKING, ClassVar

from git.objects.commit import Commit

from semantic_release.commit_parser._base import CommitParser
from semantic_release.commit_parser.conventional.options import (
    ConventionalCommitParserOptions,
)
from semantic_release.commit_parser.token import (
    ParsedCommit,
    ParsedMessageResult,
    ParseError,
    ParseResult,
)
from semantic_release.commit_parser.util import (
    breaking_re,
    deep_copy_commit,
    force_str,
    parse_paragraphs,
)
from semantic_release.enums import LevelBump
from semantic_release.errors import InvalidParserOptions
from semantic_release.helpers import sort_numerically, text_reducer

if TYPE_CHECKING:
    pass


# TODO: Remove from here, allow for user customization instead via options
# types with long names in changelog
LONG_TYPE_NAMES = {
    "build": "build system",
    "ci": "continuous integration",
    "chore": "chores",
    "docs": "documentation",
    "feat": "features",
    "fix": "bug fixes",
    "perf": "performance improvements",
    "refactor": "refactoring",
    "style": "code style",
    "test": "testing",
}


class ConventionalCommitParser(
    CommitParser[ParseResult, ConventionalCommitParserOptions]
):
    """
    A commit parser for projects conforming to the conventional commits specification.

    See https://www.conventionalcommits.org/en/v1.0.0/
    """

    # TODO: Deprecate in lieu of get_default_options()
    parser_options = ConventionalCommitParserOptions

    # GitHub & Gitea use (#123), GitLab uses (!123), and BitBucket uses (pull request #123)
    mr_selector = regexp(r"[\t ]+\((?:pull request )?(?P<mr_number>[#!]\d+)\)[\t ]*$")

    issue_selector = regexp(
        str.join(
            "",
            [
                r"^(?:clos(?:e|es|ed|ing)|fix(?:es|ed|ing)?|resolv(?:e|es|ed|ing)|implement(?:s|ed|ing)?):",
                r"[\t ]+(?P<issue_predicate>.+)[\t ]*$",
            ],
        ),
        flags=MULTILINE | IGNORECASE,
    )

    notice_selector = regexp(r"^NOTICE: (?P<notice>.+)$")

    common_commit_msg_filters: ClassVar[dict[str, tuple[Pattern[str], str]]] = {
        "typo-extra-spaces": (regexp(r"(\S)  +(\S)"), r"\1 \2"),
        "git-header-commit": (
            regexp(r"^[\t ]*commit [0-9a-f]+$\n?", flags=MULTILINE),
            "",
        ),
        "git-header-author": (
            regexp(r"^[\t ]*Author: .+$\n?", flags=MULTILINE),
            "",
        ),
        "git-header-date": (
            regexp(r"^[\t ]*Date: .+$\n?", flags=MULTILINE),
            "",
        ),
        "git-squash-heading": (
            regexp(
                r"^[\t ]*Squashed commit of the following:.*$\n?",
                flags=MULTILINE,
            ),
            "",
        ),
    }

    def __init__(self, options: ConventionalCommitParserOptions | None = None) -> None:
        super().__init__(options)

        self._logger = getLogger(
            str.join(".", [self.__module__, self.__class__.__name__])
        )

        try:
            commit_type_pattern = regexp(
                r"(?P<type>%s)" % str.join("|", self.options.allowed_tags)
            )
        except RegexError as err:
            raise InvalidParserOptions(
                str.join(
                    "\n",
                    [
                        f"Invalid options for {self.__class__.__name__}",
                        "Unable to create regular expression from configured commit-types.",
                        "Please check the configured commit-types and remove or escape any regular expression characters.",
                    ],
                )
            ) from err

        self.commit_subject = regexp(
            str.join(
                "",
                [
                    f"^{commit_type_pattern.pattern}",
                    r"(?:\((?P<scope>[^\n]+)\))?",
                    r"(?P<break>!)?:\s+",
                    r"(?P<subject>[^\n]+)",
                ],
            )
        )

        self.commit_msg_pattern = regexp(
            str.join(
                "",
                [
                    self.commit_subject.pattern,
                    r"(?:\n\n(?P<text>.+))?",  # commit body
                ],
            ),
            flags=DOTALL,
        )

        self.filters: dict[str, tuple[Pattern[str], str]] = {
            **self.common_commit_msg_filters,
            "git-squash-commit-prefix": (
                regexp(
                    str.join(
                        "",
                        [
                            r"^(?:[\t ]*[*-][\t ]+|[\t ]+)?",  # bullet points or indentation
                            commit_type_pattern.pattern + r"\b",  # prior to commit type
                        ],
                    ),
                    flags=MULTILINE,
                ),
                # move commit type to the start of the line
                r"\1",
            ),
        }

    def get_default_options(self) -> ConventionalCommitParserOptions:
        return ConventionalCommitParserOptions()

    def log_parse_error(self, commit: Commit, error: str) -> ParseError:
        self._logger.debug(error)
        return ParseError(commit, error=error)

    def commit_body_components_separator(
        self, accumulator: dict[str, list[str]], text: str
    ) -> dict[str, list[str]]:
        if (match := breaking_re.match(text)) and (brk_desc := match.group(1)):
            accumulator["breaking_descriptions"].append(brk_desc)
            return accumulator

        if (match := self.notice_selector.match(text)) and (
            notice := match.group("notice")
        ):
            accumulator["notices"].append(notice)
            return accumulator

        if match := self.issue_selector.search(text):
            # if match := self.issue_selector.search(text):
            predicate = regexp(r",? and | *[,;/& ] *").sub(
                ",", match.group("issue_predicate") or ""
            )
            # Almost all issue trackers use a number to reference an issue so
            # we use a simple regexp to validate the existence of a number which helps filter out
            # any non-issue references that don't fit our expected format
            has_number = regexp(r"\d+")
            new_issue_refs: set[str] = set(
                filter(
                    lambda issue_str, validator=has_number: validator.search(issue_str),  # type: ignore[arg-type]
                    predicate.split(","),
                )
            )
            if new_issue_refs:
                accumulator["linked_issues"] = sort_numerically(
                    set(accumulator["linked_issues"]).union(new_issue_refs)
                )
                return accumulator

        # Prevent appending duplicate descriptions
        if text not in accumulator["descriptions"]:
            accumulator["descriptions"].append(text)

        return accumulator

    def parse_message(self, message: str) -> ParsedMessageResult | None:
        return (
            self.create_parsed_message_result(match)
            if (match := self.commit_msg_pattern.match(message))
            else None
        )

    def create_parsed_message_result(
        self, match: RegexMatch[str]
    ) -> ParsedMessageResult:
        parsed_break = match.group("break")
        parsed_scope = match.group("scope") or ""
        parsed_subject = match.group("subject")
        parsed_text = match.group("text")
        parsed_type = match.group("type")

        linked_merge_request = ""
        if mr_match := self.mr_selector.search(parsed_subject):
            linked_merge_request = mr_match.group("mr_number")
            parsed_subject = self.mr_selector.sub("", parsed_subject).strip()

        body_components: dict[str, list[str]] = reduce(
            self.commit_body_components_separator,
            [
                # Insert the subject before the other paragraphs
                parsed_subject,
                *parse_paragraphs(parsed_text or ""),
            ],
            {
                "breaking_descriptions": [],
                "descriptions": [],
                "notices": [],
                "linked_issues": [],
            },
        )

        level_bump = (
            LevelBump.MAJOR
            if body_components["breaking_descriptions"] or parsed_break
            else self.options.tag_to_level.get(
                parsed_type, self.options.default_bump_level
            )
        )

        return ParsedMessageResult(
            bump=level_bump,
            type=parsed_type,
            category=LONG_TYPE_NAMES.get(parsed_type, parsed_type),
            scope=parsed_scope,
            descriptions=tuple(body_components["descriptions"]),
            breaking_descriptions=tuple(body_components["breaking_descriptions"]),
            release_notices=tuple(body_components["notices"]),
            linked_issues=tuple(body_components["linked_issues"]),
            linked_merge_request=linked_merge_request,
        )

    @staticmethod
    def is_merge_commit(commit: Commit) -> bool:
        return len(commit.parents) > 1

    def parse_commit(self, commit: Commit) -> ParseResult:
        if not (parsed_msg_result := self.parse_message(force_str(commit.message))):
            return self.log_parse_error(
                commit,
                f"Unable to parse commit message: {commit.message!r}",
            )

        return ParsedCommit.from_parsed_message_result(commit, parsed_msg_result)

    # Maybe this can be cached as an optimization, similar to how
    # mypy/pytest use their own caching directories, for very large commit
    # histories?
    # The problem is the cache likely won't be present in CI environments
    def parse(self, commit: Commit) -> ParseResult | list[ParseResult]:
        """
        Parse a commit message

        If the commit message is a squashed merge commit, it will be split into
        multiple commits, each of which will be parsed separately. Single commits
        will be returned as a list of a single ParseResult.
        """
        if self.options.ignore_merge_commits and self.is_merge_commit(commit):
            return self.log_parse_error(
                commit, "Ignoring merge commit: %s" % commit.hexsha[:8]
            )

        separate_commits: list[Commit] = (
            self.unsquash_commit(commit)
            if self.options.parse_squash_commits
            else [commit]
        )

        # Parse each commit individually if there were more than one
        parsed_commits: list[ParseResult] = list(
            map(self.parse_commit, separate_commits)
        )

        def add_linked_merge_request(
            parsed_result: ParseResult, mr_number: str
        ) -> ParseResult:
            return (
                parsed_result
                if not isinstance(parsed_result, ParsedCommit)
                else ParsedCommit(
                    **{
                        **parsed_result._asdict(),
                        "linked_merge_request": mr_number,
                    }
                )
            )

        # TODO: improve this for other VCS systems other than GitHub & BitBucket
        # Github works as the first commit in a squash merge commit has the PR number
        # appended to the first line of the commit message
        lead_commit = next(iter(parsed_commits))

        if isinstance(lead_commit, ParsedCommit) and lead_commit.linked_merge_request:
            # If the first commit has linked merge requests, assume all commits
            # are part of the same PR and add the linked merge requests to all
            # parsed commits
            parsed_commits = [
                lead_commit,
                *map(
                    lambda parsed_result, mr=lead_commit.linked_merge_request: (  # type: ignore[misc]
                        add_linked_merge_request(parsed_result, mr)
                    ),
                    parsed_commits[1:],
                ),
            ]

        elif isinstance(lead_commit, ParseError) and (
            mr_match := self.mr_selector.search(force_str(lead_commit.message))
        ):
            # Handle BitBucket Squash Merge Commits (see #1085), which have non angular commit
            # format but include the PR number in the commit subject that we want to extract
            linked_merge_request = mr_match.group("mr_number")

            # apply the linked MR to all commits
            parsed_commits = [
                add_linked_merge_request(parsed_result, linked_merge_request)
                for parsed_result in parsed_commits
            ]

        return parsed_commits

    def unsquash_commit(self, commit: Commit) -> list[Commit]:
        # GitHub EXAMPLE:
        # feat(changelog): add autofit_text_width filter to template environment (#1062)
        #
        # This change adds an equivalent style formatter that can apply a text alignment
        # to a maximum width and also maintain an indent over paragraphs of text
        #
        # * docs(changelog-templates): add definition & usage of autofit_text_width template filter
        #
        # * test(changelog-context): add test cases to check autofit_text_width filter use
        #
        # `git merge --squash` EXAMPLE:
        # Squashed commit of the following:
        #
        # commit 63ec09b9e844e616dcaa7bae35a0b66671b59fbb
        # Author: codejedi365 <codejedi365@gmail.com>
        # Date:   Sun Oct 13 12:05:23 2024 -0600
        #
        #     feat(release-config): some commit subject
        #

        # Return a list of artificial commits (each with a single commit message)
        return [
            # create a artificial commit object (copy of original but with modified message)
            Commit(
                **{
                    **deep_copy_commit(commit),
                    "message": commit_msg,
                }
            )
            for commit_msg in self.unsquash_commit_message(force_str(commit.message))
        ] or [commit]

    def unsquash_commit_message(self, message: str) -> list[str]:
        normalized_message = message.replace("\r", "").strip()

        # split by obvious separate commits (applies to manual git squash merges)
        obvious_squashed_commits = self.filters["git-header-commit"][0].split(
            normalized_message
        )

        separate_commit_msgs: list[str] = reduce(
            lambda all_msgs, msgs: all_msgs + msgs,
            map(self._find_squashed_commits_in_str, obvious_squashed_commits),
            [],
        )

        return list(filter(None, separate_commit_msgs))

    def _find_squashed_commits_in_str(self, message: str) -> list[str]:
        separate_commit_msgs: list[str] = []
        current_msg = ""

        for paragraph in filter(None, message.strip().split("\n\n")):
            # Apply filters to normalize the paragraph
            clean_paragraph = reduce(text_reducer, self.filters.values(), paragraph)

            # remove any filtered (and now empty) paragraphs (ie. the git headers)
            if not clean_paragraph.strip():
                continue

            # Check if the paragraph is the start of a new conventional commit
            # Note: that we check that the subject has more than one word to differentiate from
            # a closing footer (e.g. "fix: #123", or "fix: ABC-123")
            if (match := self.commit_subject.search(clean_paragraph)) and len(
                match.group("subject").split(" ")
            ) > 1:
                # Since we found the start of the new commit, store any previous commit
                # message separately and start the new commit message
                if current_msg:
                    separate_commit_msgs.append(current_msg)

                current_msg = clean_paragraph
                continue

            if not separate_commit_msgs and not current_msg:
                # if there are no separate commit messages and no current message
                # then this is the first commit message
                current_msg = dedent(clean_paragraph)
                continue

            # append the paragraph as part of the previous commit message
            if current_msg:
                current_msg += f"\n\n{dedent(clean_paragraph)}"

            # else: drop the paragraph
            continue

        return [*separate_commit_msgs, current_msg]



from __future__ import annotations

import os
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING
from unittest import mock

import pytest

# NOTE: use backport with newer API to support 3.7
from importlib_resources import files

import semantic_release
from semantic_release.cli.changelog_writer import generate_release_notes
from semantic_release.commit_parser.token import ParsedCommit
from semantic_release.hvcs import Bitbucket, Gitea, Github, Gitlab

if TYPE_CHECKING:
    from semantic_release.changelog.release_history import ReleaseHistory

    from tests.fixtures.example_project import ExProjectDir


@pytest.fixture(scope="module")
def release_notes_template() -> str:
    """Retrieve the semantic-release default release notes template."""
    version_notes_template = files(semantic_release.__name__).joinpath(
        Path("data", "templates", "conventional", "md", ".release_notes.md.j2")
    )
    return version_notes_template.read_text(encoding="utf-8")


@pytest.mark.parametrize("mask_initial_release", [True, False])
@pytest.mark.parametrize("license_name", ["", "MIT"])
@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_release_notes_template(
    example_git_https_url: str,
    hvcs_client: type[Github | Gitlab | Gitea | Bitbucket],
    license_name: str,
    artificial_release_history: ReleaseHistory,
    mask_initial_release: bool,
    today_date_str: str,
):
    """
    Unit test goal: just make sure it renders the release notes template without error.

    Scenarios are better suited for all the variations (commit types).
    """
    released_versions = iter(artificial_release_history.released.keys())
    version = next(released_versions)
    prev_version = next(released_versions)
    hvcs = hvcs_client(example_git_https_url)
    release = artificial_release_history.released[version]

    feat_commit_obj = release["elements"]["feature"][0]
    fix_commit_obj_1 = release["elements"]["fix"][0]
    fix_commit_obj_2 = release["elements"]["fix"][1]
    fix_commit_obj_3 = release["elements"]["fix"][2]
    assert isinstance(feat_commit_obj, ParsedCommit)
    assert isinstance(fix_commit_obj_1, ParsedCommit)
    assert isinstance(fix_commit_obj_2, ParsedCommit)
    assert isinstance(fix_commit_obj_3, ParsedCommit)

    feat_commit_url = hvcs.commit_hash_url(feat_commit_obj.commit.hexsha)
    feat_description = str.join("\n", feat_commit_obj.descriptions)

    fix_commit_1_url = hvcs.commit_hash_url(fix_commit_obj_1.commit.hexsha)
    fix_commit_1_description = str.join("\n", fix_commit_obj_1.descriptions)

    fix_commit_2_url = hvcs.commit_hash_url(fix_commit_obj_2.commit.hexsha)
    fix_commit_2_description = str.join("\n", fix_commit_obj_2.descriptions)

    fix_commit_3_url = hvcs.commit_hash_url(fix_commit_obj_3.commit.hexsha)
    fix_commit_3_description = str.join("\n", fix_commit_obj_3.descriptions)

    expected_content = str.join(
        os.linesep,
        [
            f"## v{version} ({today_date_str})",
            *(
                [""]
                if not license_name
                else [
                    "",
                    f"_This release is published under the {license_name} License._",
                    "",
                ]
            ),
            "### Feature",
            "",
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{feat_commit_obj.scope}**: " if feat_commit_obj.scope else ""
                ),
                commit_desc=feat_description.capitalize(),
                short_hash=feat_commit_obj.commit.hexsha[:7],
                url=feat_commit_url,
            ),
            "",
            "### Fix",
            "",
            # Commit 2 is first because it has no scope
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{fix_commit_obj_2.scope}**: " if fix_commit_obj_2.scope else ""
                ),
                commit_desc=fix_commit_2_description.capitalize(),
                short_hash=fix_commit_obj_2.commit.hexsha[:7],
                url=fix_commit_2_url,
            ),
            "",
            # Commit 3 is second because it starts with an A even though it has the same scope as 1
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{fix_commit_obj_3.scope}**: " if fix_commit_obj_3.scope else ""
                ),
                commit_desc=fix_commit_3_description.capitalize(),
                short_hash=fix_commit_obj_3.commit.hexsha[:7],
                url=fix_commit_3_url,
            ),
            "",
            # Commit 1 is last
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{fix_commit_obj_1.scope}**: " if fix_commit_obj_1.scope else ""
                ),
                commit_desc=fix_commit_1_description.capitalize(),
                short_hash=fix_commit_obj_1.commit.hexsha[:7],
                url=fix_commit_1_url,
            ),
            "",
        ],
    )

    if not isinstance(hvcs, Gitea):
        expected_content += str.join(
            os.linesep,
            [
                "",
                "---",
                "",
                "**Detailed Changes**: [{prev_version}...{new_version}]({version_compare_url})".format(
                    prev_version=prev_version.as_tag(),
                    new_version=version.as_tag(),
                    version_compare_url=hvcs.compare_url(
                        prev_version.as_tag(), version.as_tag()
                    ),
                ),
                "",
            ],
        )

    actual_content = generate_release_notes(
        hvcs_client=hvcs_client(remote_url=example_git_https_url),
        release=release,
        template_dir=Path(""),
        history=artificial_release_history,
        style="conventional",
        mask_initial_release=mask_initial_release,
        license_name=license_name,
    )

    assert expected_content == actual_content


@pytest.mark.parametrize("mask_initial_release", [True, False])
@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_release_notes_template_w_a_brk_description(
    example_git_https_url: str,
    hvcs_client: type[Github | Gitlab | Gitea | Bitbucket],
    release_history_w_brk_change: ReleaseHistory,
    mask_initial_release: bool,
    today_date_str: str,
):
    """
    Unit test goal: just make sure it renders the release notes template without error.

    Scenarios are better suited for all the variations (commit types).
    """
    released_versions = iter(release_history_w_brk_change.released.keys())
    version = next(released_versions)
    prev_version = next(released_versions)
    hvcs = hvcs_client(example_git_https_url)
    release = release_history_w_brk_change.released[version]

    brk_fix_commit_obj = next(iter(release["elements"].values()))[0]
    assert isinstance(brk_fix_commit_obj, ParsedCommit)

    brk_fix_commit_url = hvcs.commit_hash_url(brk_fix_commit_obj.commit.hexsha)
    brk_fix_description = str.join("\n", brk_fix_commit_obj.descriptions)
    brk_fix_brking_description = str.join(
        "\n", brk_fix_commit_obj.breaking_descriptions
    )

    expected_content = str.join(
        os.linesep,
        [
            f"## v{version} ({today_date_str})",
            "",
            "### Bug Fixes",
            "",
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{brk_fix_commit_obj.scope}**: "
                    if brk_fix_commit_obj.scope
                    else ""
                ),
                commit_desc=brk_fix_description.capitalize(),
                short_hash=brk_fix_commit_obj.commit.hexsha[:7],
                url=brk_fix_commit_url,
            ),
            "",
            "### Breaking Changes",
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{brk_fix_commit_obj.scope}**: "
                    if brk_fix_commit_obj.scope
                    else ""
                ),
                change_desc=brk_fix_brking_description.capitalize(),
            ),
            "",
        ],
    )

    if not isinstance(hvcs, Gitea):
        expected_content += str.join(
            os.linesep,
            [
                "",
                "---",
                "",
                "**Detailed Changes**: [{prev_version}...{new_version}]({version_compare_url})".format(
                    prev_version=prev_version.as_tag(),
                    new_version=version.as_tag(),
                    version_compare_url=hvcs.compare_url(
                        prev_version.as_tag(), version.as_tag()
                    ),
                ),
                "",
            ],
        )

    actual_content = generate_release_notes(
        hvcs_client=hvcs_client(remote_url=example_git_https_url),
        release=release,
        template_dir=Path(""),
        history=release_history_w_brk_change,
        style="conventional",
        mask_initial_release=mask_initial_release,
    )

    assert expected_content == actual_content


@pytest.mark.parametrize("mask_initial_release", [True, False])
@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_release_notes_template_w_multiple_brk_changes(
    example_git_https_url: str,
    hvcs_client: type[Github | Gitlab | Gitea | Bitbucket],
    release_history_w_multiple_brk_changes: ReleaseHistory,
    mask_initial_release: bool,
    today_date_str: str,
):
    """
    Unit test goal: just make sure it renders the release notes template without error.

    Scenarios are better suited for all the variations (commit types).
    """
    released_versions = iter(release_history_w_multiple_brk_changes.released.keys())
    version = next(released_versions)
    prev_version = next(released_versions)
    hvcs = hvcs_client(example_git_https_url)
    release = release_history_w_multiple_brk_changes.released[version]

    brk_fix_commit_obj = release["elements"]["Bug Fixes"][0]
    brk_feat_commit_obj = release["elements"]["Features"][0]
    assert isinstance(brk_fix_commit_obj, ParsedCommit)
    assert isinstance(brk_feat_commit_obj, ParsedCommit)

    brk_fix_commit_url = hvcs.commit_hash_url(brk_fix_commit_obj.commit.hexsha)
    brk_fix_description = str.join("\n", brk_fix_commit_obj.descriptions)
    brk_fix_brking_description = str.join(
        "\n", brk_fix_commit_obj.breaking_descriptions
    )

    brk_feat_commit_url = hvcs.commit_hash_url(brk_feat_commit_obj.commit.hexsha)
    brk_feat_description = str.join("\n", brk_feat_commit_obj.descriptions)
    brk_feat_brking_description = str.join(
        "\n", brk_feat_commit_obj.breaking_descriptions
    )

    expected_content = str.join(
        os.linesep,
        [
            f"## v{version} ({today_date_str})",
            "",
            "### Bug Fixes",
            "",
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{brk_fix_commit_obj.scope}**: "
                    if brk_fix_commit_obj.scope
                    else ""
                ),
                commit_desc=brk_fix_description.capitalize(),
                short_hash=brk_fix_commit_obj.commit.hexsha[:7],
                url=brk_fix_commit_url,
            ),
            "",
            "### Features",
            "",
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{brk_feat_commit_obj.scope}**: "
                    if brk_feat_commit_obj.scope
                    else ""
                ),
                commit_desc=brk_feat_description.capitalize(),
                short_hash=brk_feat_commit_obj.commit.hexsha[:7],
                url=brk_feat_commit_url,
            ),
            "",
            "### Breaking Changes",
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{brk_feat_commit_obj.scope}**: "
                    if brk_feat_commit_obj.scope
                    else ""
                ),
                change_desc=brk_feat_brking_description.capitalize(),
            ),
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{brk_fix_commit_obj.scope}**: "
                    if brk_fix_commit_obj.scope
                    else ""
                ),
                change_desc=brk_fix_brking_description.capitalize(),
            ),
            "",
        ],
    )

    if not isinstance(hvcs, Gitea):
        expected_content += str.join(
            os.linesep,
            [
                "",
                "---",
                "",
                "**Detailed Changes**: [{prev_version}...{new_version}]({version_compare_url})".format(
                    prev_version=prev_version.as_tag(),
                    new_version=version.as_tag(),
                    version_compare_url=hvcs.compare_url(
                        prev_version.as_tag(), version.as_tag()
                    ),
                ),
                "",
            ],
        )

    actual_content = generate_release_notes(
        hvcs_client=hvcs_client(remote_url=example_git_https_url),
        release=release,
        template_dir=Path(""),
        history=release_history_w_multiple_brk_changes,
        style="conventional",
        mask_initial_release=mask_initial_release,
    )

    assert expected_content == actual_content


@pytest.mark.parametrize("license_name", ["", "MIT"])
@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_release_notes_template_first_release_masked(
    example_git_https_url: str,
    hvcs_client: type[Bitbucket | Gitea | Github | Gitlab],
    license_name: str,
    single_release_history: ReleaseHistory,
    today_date_str: str,
):
    """
    Unit test goal: just make sure it renders the release notes template without error.

    Scenarios are better suited for all the variations (commit types).
    """
    hvcs = hvcs_client(example_git_https_url)
    version = list(single_release_history.released.keys())[-1]
    release = single_release_history.released[version]

    expected_content = str.join(
        os.linesep,
        [
            f"## v{version} ({today_date_str})",
            *(
                [""]
                if not license_name
                else [
                    "",
                    f"_This release is published under the {license_name} License._",
                    "",
                ]
            ),
            "- Initial Release",
            "",
        ],
    )

    actual_content = generate_release_notes(
        hvcs_client=hvcs,
        release=release,
        template_dir=Path(""),
        history=single_release_history,
        style="conventional",
        mask_initial_release=True,
        license_name=license_name,
    )

    assert expected_content == actual_content


@pytest.mark.parametrize("license_name", ["", "MIT"])
@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_release_notes_template_first_release_unmasked(
    example_git_https_url: str,
    hvcs_client: type[Bitbucket | Gitea | Github | Gitlab],
    license_name: str,
    single_release_history: ReleaseHistory,
    today_date_str: str,
):
    """
    Unit test goal: just make sure it renders the release notes template without error.

    Scenarios are better suited for all the variations (commit types).
    """
    hvcs = hvcs_client(example_git_https_url)
    version = list(single_release_history.released.keys())[-1]
    release = single_release_history.released[version]

    feat_commit_obj = release["elements"]["feature"][0]
    assert isinstance(feat_commit_obj, ParsedCommit)

    feat_commit_url = hvcs.commit_hash_url(feat_commit_obj.commit.hexsha)
    feat_description = str.join("\n", feat_commit_obj.descriptions)

    expected_content = str.join(
        os.linesep,
        [
            f"## v{version} ({today_date_str})",
            *(
                [""]
                if not license_name
                else [
                    "",
                    f"_This release is published under the {license_name} License._",
                    "",
                ]
            ),
            "### Feature",
            "",
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{feat_commit_obj.scope}**: " if feat_commit_obj.scope else ""
                ),
                commit_desc=feat_description.capitalize(),
                short_hash=feat_commit_obj.commit.hexsha[:7],
                url=feat_commit_url,
            ),
            "",
        ],
    )

    actual_content = generate_release_notes(
        hvcs_client=hvcs,
        release=release,
        template_dir=Path(""),
        history=single_release_history,
        style="conventional",
        mask_initial_release=False,
        license_name=license_name,
    )

    assert expected_content == actual_content


def test_release_notes_context_sort_numerically_filter(
    example_git_https_url: str,
    single_release_history: ReleaseHistory,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
):
    version = list(single_release_history.released.keys())[-1]
    release = single_release_history.released[version]

    example_project_dir.joinpath(".release_notes.md.j2").write_text(
        dedent(
            """\
            {{ [
                    ".. _#5: link",
                    ".. _PR#3: link",
                    ".. _PR#10: link",
                    ".. _#100: link"
                ] | sort_numerically | join("\\n")
            }}
            """
        )
    )

    expected_content = str.join(
        os.linesep,
        dedent(
            """\
            .. _#5: link
            .. _#100: link
            .. _PR#3: link
            .. _PR#10: link
            """
        ).split("\n"),
    )

    actual_content = generate_release_notes(
        hvcs_client=Github(remote_url=example_git_https_url),
        release=release,
        template_dir=example_project_dir,
        history=single_release_history,
        style="conventional",
        mask_initial_release=False,
    )

    assert expected_content == actual_content


def test_release_notes_context_sort_numerically_filter_reversed(
    example_git_https_url: str,
    single_release_history: ReleaseHistory,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
):
    version = list(single_release_history.released.keys())[-1]
    release = single_release_history.released[version]

    example_project_dir.joinpath(".release_notes.md.j2").write_text(
        dedent(
            """\
            {{ [
                    ".. _#5: link",
                    ".. _PR#3: link",
                    ".. _PR#10: link",
                    ".. _#100: link"
                ] | sort_numerically(reverse=True) | join("\\n")
            }}
            """
        )
    )

    expected_content = str.join(
        os.linesep,
        dedent(
            """\
            .. _#100: link
            .. _#5: link
            .. _PR#10: link
            .. _PR#3: link
            """
        ).split("\n"),
    )

    actual_content = generate_release_notes(
        hvcs_client=Github(remote_url=example_git_https_url),
        release=release,
        template_dir=example_project_dir,
        history=single_release_history,
        style="conventional",
        mask_initial_release=False,
    )

    assert expected_content == actual_content


def test_release_notes_context_pypi_url_filter(
    example_git_https_url: str,
    single_release_history: ReleaseHistory,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
):
    version = list(single_release_history.released.keys())[-1]
    release = single_release_history.released[version]

    example_project_dir.joinpath(".release_notes.md.j2").write_text(
        """{{ "example-package" | create_pypi_url }}"""
    )

    expected_content = f"https://pypi.org/project/example-package{os.linesep}"

    actual_content = generate_release_notes(
        hvcs_client=Github(remote_url=example_git_https_url),
        release=release,
        template_dir=example_project_dir,
        history=single_release_history,
        style="conventional",
        mask_initial_release=False,
    )

    assert expected_content == actual_content


def test_release_notes_context_pypi_url_filter_tagged(
    example_git_https_url: str,
    single_release_history: ReleaseHistory,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
):
    version = list(single_release_history.released.keys())[-1]
    release = single_release_history.released[version]

    example_project_dir.joinpath(".release_notes.md.j2").write_text(
        """{{ "example-package" | create_pypi_url(release.version | string) }}"""
    )

    expected_content = f"https://pypi.org/project/example-package/{version}{os.linesep}"

    actual_content = generate_release_notes(
        hvcs_client=Github(remote_url=example_git_https_url),
        release=release,
        template_dir=example_project_dir,
        history=single_release_history,
        style="conventional",
        mask_initial_release=False,
    )

    assert expected_content == actual_content


@pytest.mark.parametrize("hvcs_client_class", [Github, Gitlab, Gitea])
def test_release_notes_context_release_url_filter(
    example_git_https_url: str,
    hvcs_client_class: type[Github | Gitlab | Gitea],
    single_release_history: ReleaseHistory,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
):
    version = list(single_release_history.released.keys())[-1]
    release = single_release_history.released[version]

    example_project_dir.joinpath(".release_notes.md.j2").write_text(
        dedent(
            """\
            {{
                "[%s](%s)" | format(
                    release.version.as_tag(),
                    release.version.as_tag() | create_release_url,
                )
            }}
            """
        )
    )

    with mock.patch.dict(os.environ, {}, clear=True):
        hvcs_client = hvcs_client_class(remote_url=example_git_https_url)

        expected_content = dedent(
            f"""\
            [{version.as_tag()}]({hvcs_client.create_release_url(version.as_tag())})
            """
        )

        actual_content = generate_release_notes(
            hvcs_client=hvcs_client,
            release=release,
            template_dir=example_project_dir,
            history=single_release_history,
            style="conventional",
            mask_initial_release=False,
        )

    assert expected_content == actual_content


@pytest.mark.parametrize("hvcs_client_class", [Github, Gitlab, Gitea, Bitbucket])
def test_release_notes_context_format_w_official_name_filter(
    example_git_https_url: str,
    hvcs_client_class: type[Github | Gitlab | Gitea],
    single_release_history: ReleaseHistory,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
):
    version = list(single_release_history.released.keys())[-1]
    release = single_release_history.released[version]

    example_project_dir.joinpath(".release_notes.md.j2").write_text(
        dedent(
            """\
            {{ "%s" | format_w_official_vcs_name }}
            {{ "{}" | format_w_official_vcs_name }}
            {{ "{vcs_name}" | format_w_official_vcs_name }}
            """
        )
    )

    with mock.patch.dict(os.environ, {}, clear=True):
        hvcs_client = hvcs_client_class(remote_url=example_git_https_url)
        expected_content = dedent(
            f"""\
            {hvcs_client.OFFICIAL_NAME}
            {hvcs_client.OFFICIAL_NAME}
            {hvcs_client.OFFICIAL_NAME}
            """
        )

        actual_content = generate_release_notes(
            hvcs_client=hvcs_client,
            release=release,
            template_dir=example_project_dir,
            history=single_release_history,
            style="conventional",
            mask_initial_release=False,
        )

    assert expected_content == actual_content


@pytest.mark.parametrize("mask_initial_release", [True, False])
@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_release_notes_template_w_a_notice(
    example_git_https_url: str,
    hvcs_client: type[Github | Gitlab | Gitea | Bitbucket],
    release_history_w_a_notice: ReleaseHistory,
    mask_initial_release: bool,
    today_date_str: str,
):
    """
    Unit test goal: just make sure it renders the release notes template without error.

    Scenarios are better suited for all the variations (commit types).
    """
    released_versions = iter(release_history_w_a_notice.released.keys())
    version = next(released_versions)
    prev_version = next(released_versions)
    hvcs = hvcs_client(example_git_https_url)
    release = release_history_w_a_notice.released[version]

    notice_commit_obj = next(iter(release["elements"].values()))[0]
    assert isinstance(notice_commit_obj, ParsedCommit)

    notice_commit_url = hvcs.commit_hash_url(notice_commit_obj.commit.hexsha)
    notice_commit_description = str.join("\n", notice_commit_obj.descriptions)
    notice_description = str.join("\n", notice_commit_obj.release_notices)

    expected_content = str.join(
        os.linesep,
        [
            f"## v{version} ({today_date_str})",
            "",
            "### Refactoring",
            "",
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{notice_commit_obj.scope}**: "
                    if notice_commit_obj.scope
                    else ""
                ),
                commit_desc=notice_commit_description.capitalize().rstrip(),
                short_hash=notice_commit_obj.commit.hexsha[:7],
                url=notice_commit_url,
            ),
            "",
            "### Additional Release Information",
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{notice_commit_obj.scope}**: "
                    if notice_commit_obj.scope
                    else ""
                ),
                change_desc=notice_description.capitalize().rstrip(),
            ),
            "",
        ],
    )

    if not isinstance(hvcs, Gitea):
        expected_content += str.join(
            os.linesep,
            [
                "",
                "---",
                "",
                "**Detailed Changes**: [{prev_version}...{new_version}]({version_compare_url})".format(
                    prev_version=prev_version.as_tag(),
                    new_version=version.as_tag(),
                    version_compare_url=hvcs.compare_url(
                        prev_version.as_tag(), version.as_tag()
                    ),
                ),
                "",
            ],
        )

    actual_content = generate_release_notes(
        hvcs_client=hvcs_client(remote_url=example_git_https_url),
        release=release,
        template_dir=Path(""),
        history=release_history_w_a_notice,
        style="conventional",
        mask_initial_release=mask_initial_release,
    )

    assert expected_content == actual_content


@pytest.mark.parametrize("mask_initial_release", [True, False])
@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_release_notes_template_w_a_notice_n_brk_change(
    example_git_https_url: str,
    hvcs_client: type[Github | Gitlab | Gitea | Bitbucket],
    release_history_w_notice_n_brk_change: ReleaseHistory,
    mask_initial_release: bool,
    today_date_str: str,
):
    """
    Unit test goal: just make sure it renders the release notes template without error.

    Scenarios are better suited for all the variations (commit types).
    """
    released_versions = iter(release_history_w_notice_n_brk_change.released.keys())
    version = next(released_versions)
    prev_version = next(released_versions)
    hvcs = hvcs_client(example_git_https_url)
    release = release_history_w_notice_n_brk_change.released[version]

    brk_fix_commit_obj = release["elements"]["Bug Fixes"][0]
    notice_commit_obj = release["elements"]["Refactoring"][0]
    assert isinstance(brk_fix_commit_obj, ParsedCommit)
    assert isinstance(notice_commit_obj, ParsedCommit)

    brk_fix_commit_url = hvcs.commit_hash_url(brk_fix_commit_obj.commit.hexsha)
    brk_fix_description = str.join("\n", brk_fix_commit_obj.descriptions)
    brk_fix_brking_description = str.join(
        "\n", brk_fix_commit_obj.breaking_descriptions
    )

    notice_commit_url = hvcs.commit_hash_url(notice_commit_obj.commit.hexsha)
    notice_commit_description = str.join("\n", notice_commit_obj.descriptions)
    notice_description = str.join("\n", notice_commit_obj.release_notices)

    expected_content = str.join(
        os.linesep,
        [
            f"## v{version} ({today_date_str})",
            "",
            "### Bug Fixes",
            "",
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{brk_fix_commit_obj.scope}**: "
                    if brk_fix_commit_obj.scope
                    else ""
                ),
                commit_desc=brk_fix_description.capitalize().rstrip(),
                short_hash=brk_fix_commit_obj.commit.hexsha[:7],
                url=brk_fix_commit_url,
            ),
            "",
            "### Refactoring",
            "",
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{notice_commit_obj.scope}**: "
                    if notice_commit_obj.scope
                    else ""
                ),
                commit_desc=notice_commit_description.capitalize().rstrip(),
                short_hash=notice_commit_obj.commit.hexsha[:7],
                url=notice_commit_url,
            ),
            "",
            "### Breaking Changes",
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{brk_fix_commit_obj.scope}**: "
                    if brk_fix_commit_obj.scope
                    else ""
                ),
                change_desc=brk_fix_brking_description.capitalize().rstrip(),
            ),
            "",
            "### Additional Release Information",
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{notice_commit_obj.scope}**: "
                    if notice_commit_obj.scope
                    else ""
                ),
                change_desc=notice_description.capitalize().rstrip(),
            ),
            "",
        ],
    )

    if not isinstance(hvcs, Gitea):
        expected_content += str.join(
            os.linesep,
            [
                "",
                "---",
                "",
                "**Detailed Changes**: [{prev_version}...{new_version}]({version_compare_url})".format(
                    prev_version=prev_version.as_tag(),
                    new_version=version.as_tag(),
                    version_compare_url=hvcs.compare_url(
                        prev_version.as_tag(), version.as_tag()
                    ),
                ),
                "",
            ],
        )

    actual_content = generate_release_notes(
        hvcs_client=hvcs_client(remote_url=example_git_https_url),
        release=release,
        template_dir=Path(""),
        history=release_history_w_notice_n_brk_change,
        style="conventional",
        mask_initial_release=mask_initial_release,
    )

    assert expected_content == actual_content


@pytest.mark.parametrize("mask_initial_release", [True, False])
@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_release_notes_template_w_multiple_notices(
    example_git_https_url: str,
    hvcs_client: type[Github | Gitlab | Gitea | Bitbucket],
    release_history_w_multiple_notices: ReleaseHistory,
    mask_initial_release: bool,
    today_date_str: str,
):
    """
    Unit test goal: just make sure it renders the release notes template without error.

    Scenarios are better suited for all the variations (commit types).
    """
    released_versions = iter(release_history_w_multiple_notices.released.keys())
    version = next(released_versions)
    prev_version = next(released_versions)
    hvcs = hvcs_client(example_git_https_url)
    release = release_history_w_multiple_notices.released[version]

    feat_commit_obj = release["elements"]["Features"][0]
    refactor_commit_obj = release["elements"]["Refactoring"][0]
    assert isinstance(refactor_commit_obj, ParsedCommit)
    assert isinstance(feat_commit_obj, ParsedCommit)

    refactor_commit_url = hvcs.commit_hash_url(refactor_commit_obj.commit.hexsha)
    refactor_commit_desc = str.join("\n", refactor_commit_obj.descriptions)
    refactor_commit_notice_desc = str.join("\n", refactor_commit_obj.release_notices)

    feat_commit_url = hvcs.commit_hash_url(feat_commit_obj.commit.hexsha)
    feat_description = str.join("\n", feat_commit_obj.descriptions)
    feat_commit_notice_desc = str.join("\n", feat_commit_obj.release_notices)

    expected_content = str.join(
        os.linesep,
        [
            f"## v{version} ({today_date_str})",
            "",
            "### Features",
            "",
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{feat_commit_obj.scope}**: " if feat_commit_obj.scope else ""
                ),
                commit_desc=feat_description.capitalize().rstrip(),
                short_hash=feat_commit_obj.commit.hexsha[:7],
                url=feat_commit_url,
            ),
            "",
            "### Refactoring",
            "",
            "- {commit_scope}{commit_desc} ([`{short_hash}`]({url}))".format(
                commit_scope=(
                    f"**{refactor_commit_obj.scope}**: "
                    if refactor_commit_obj.scope
                    else ""
                ),
                commit_desc=refactor_commit_desc.capitalize().rstrip(),
                short_hash=refactor_commit_obj.commit.hexsha[:7],
                url=refactor_commit_url,
            ),
            "",
            "### Additional Release Information",
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{refactor_commit_obj.scope}**: "
                    if refactor_commit_obj.scope
                    else ""
                ),
                change_desc=refactor_commit_notice_desc.capitalize().rstrip(),
            ),
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{feat_commit_obj.scope}**: " if feat_commit_obj.scope else ""
                ),
                change_desc=feat_commit_notice_desc.capitalize().rstrip(),
            ),
            "",
        ],
    )

    if not isinstance(hvcs, Gitea):
        expected_content += str.join(
            os.linesep,
            [
                "",
                "---",
                "",
                "**Detailed Changes**: [{prev_version}...{new_version}]({version_compare_url})".format(
                    prev_version=prev_version.as_tag(),
                    new_version=version.as_tag(),
                    version_compare_url=hvcs.compare_url(
                        prev_version.as_tag(), version.as_tag()
                    ),
                ),
                "",
            ],
        )

    actual_content = generate_release_notes(
        hvcs_client=hvcs_client(remote_url=example_git_https_url),
        release=release,
        template_dir=Path(""),
        history=release_history_w_multiple_notices,
        style="conventional",
        mask_initial_release=mask_initial_release,
    )

    assert expected_content == actual_content



from __future__ import annotations

import json

import click
import tomlkit

from semantic_release.cli.config import RawConfig


@click.command(
    short_help="Generate semantic-release's default configuration",
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
)
@click.option(
    "-f",
    "--format",
    "fmt",
    type=click.Choice(["toml", "json"], case_sensitive=False),
    default="toml",
    help="format for the config to be generated",
)
@click.option(
    "--pyproject",
    "is_pyproject_toml",
    is_flag=True,
    help=(
        "Add TOML configuration under 'tool.semantic_release' instead of "
        "'semantic_release'"
    ),
)
def generate_config(fmt: str = "toml", is_pyproject_toml: bool = False) -> None:
    """
    Generate default configuration for semantic-release, to help you get started
    quickly. You can inspect the defaults, write to a file and then edit according to
    your needs. For example, to append the default configuration to your pyproject.toml
    file, you can use the following command:

        semantic-release generate-config --pyproject >> pyproject.toml
    """
    # due to possible IntEnum values (which are not supported by tomlkit.dumps, see sdispater/tomlkit#237),
    # we must ensure the transformation of the model to a dict uses json serializable values
    config = RawConfig().model_dump(mode="json", exclude_none=True)

    config_dct = {"semantic_release": config}
    if is_pyproject_toml and fmt == "toml":
        config_dct = {"tool": config_dct}

    if fmt == "toml":
        click.echo(tomlkit.dumps(config_dct))

    elif fmt == "json":
        click.echo(json.dumps(config_dct, indent=4))



from __future__ import annotations

import itertools
import os
from typing import TYPE_CHECKING

import pytest

from semantic_release.changelog.template import environment, recursive_render

if TYPE_CHECKING:
    from pathlib import Path

    from tests.fixtures.example_project import ExProjectDir


NORMAL_TEMPLATE_SRC = """---
content:
    - a string
    - ["a nested list"]
vars:
    # a comment
    hello: {{ "world" | upper }}

"""

NORMAL_TEMPLATE_RENDERED = """---
content:
    - a string
    - ["a nested list"]
vars:
    # a comment
    hello: WORLD
"""

PLAINTEXT_FILE_CONTENT = """
I should not be rendered as a template.
{{ "this string should be untouched" | upper }}
"""


def _strip_trailing_j2(path: Path) -> Path:
    if path.name.endswith(".j2"):
        return path.with_name(path.name[:-3])
    return path


@pytest.fixture
def normal_template(example_project_template_dir: Path) -> Path:
    template = example_project_template_dir / "normal.yaml.j2"
    template.parent.mkdir(parents=True, exist_ok=True)
    template.write_text(NORMAL_TEMPLATE_SRC)
    return template


@pytest.fixture
def long_directory_path(example_project_template_dir: Path) -> Path:
    # NOTE: fixture enables using Path rather than
    # constant string, so no issue with / vs \ on Windows
    return example_project_template_dir / "long" / "dir" / "path"


@pytest.fixture
def deeply_nested_file(long_directory_path: Path) -> Path:
    file = long_directory_path / "buried.txt"
    file.parent.mkdir(parents=True, exist_ok=True)
    file.write_text(PLAINTEXT_FILE_CONTENT)
    return file


@pytest.fixture
def hidden_file(example_project_template_dir: Path) -> Path:
    file = example_project_template_dir / ".hidden"
    file.parent.mkdir(parents=True, exist_ok=True)
    file.write_text("I shouldn't be present")
    return file


@pytest.fixture
def directory_path_with_hidden_subfolder(example_project_template_dir: Path) -> Path:
    return example_project_template_dir / "path" / ".subfolder" / "hidden"


@pytest.fixture
def excluded_file(directory_path_with_hidden_subfolder: Path) -> Path:
    file = directory_path_with_hidden_subfolder / "excluded.txt"
    file.parent.mkdir(parents=True, exist_ok=True)
    file.write_text("I shouldn't be present")
    return file


@pytest.mark.usefixtures(excluded_file.__name__)
def test_recursive_render(
    init_example_project: None,
    example_project_dir: Path,
    example_project_template_dir: Path,
    normal_template,
    deeply_nested_file,
    hidden_file,
):
    tmpl_dir = str(example_project_template_dir.resolve())
    env = environment(template_dir=tmpl_dir)

    preexisting_paths = set(example_project_dir.rglob("**/*"))

    recursive_render(
        template_dir=example_project_template_dir.resolve(),
        environment=env,
        _root_dir=str(example_project_dir.resolve()),
    )
    rendered_normal_template = _strip_trailing_j2(
        example_project_dir / normal_template.relative_to(example_project_template_dir)
    )
    assert rendered_normal_template.exists()
    assert rendered_normal_template.read_text() == NORMAL_TEMPLATE_RENDERED

    rendered_deeply_nested = example_project_dir / deeply_nested_file.relative_to(
        example_project_template_dir
    )
    assert rendered_deeply_nested.exists()
    assert rendered_deeply_nested.read_text() == PLAINTEXT_FILE_CONTENT

    rendered_hidden = example_project_dir / hidden_file.relative_to(
        example_project_template_dir
    )
    assert not rendered_hidden.exists()

    assert not (example_project_dir / "path").exists()

    assert set(example_project_dir.rglob("**/*")) == preexisting_paths.union(
        example_project_dir / p
        for t in (
            rendered_normal_template,
            rendered_deeply_nested,
        )
        for p in itertools.accumulate(
            t.relative_to(example_project_dir).parts, func=lambda *a: os.sep.join(a)
        )
    )


@pytest.fixture
def dotfolder_template_dir(example_project_dir: ExProjectDir) -> Path:
    return example_project_dir / ".templates/.psr-templates"


@pytest.fixture
def dotfolder_template(
    init_example_project: None, dotfolder_template_dir: Path
) -> Path:
    tmpl = dotfolder_template_dir / "template.txt"
    tmpl.parent.mkdir(parents=True, exist_ok=True)
    tmpl.write_text("I am a template")
    return tmpl


def test_recursive_render_with_top_level_dotfolder(
    init_example_project: None,
    example_project_dir: ExProjectDir,
    dotfolder_template: Path,
    dotfolder_template_dir: Path,
):
    preexisting_paths = set(example_project_dir.rglob("**/*"))
    env = environment(template_dir=dotfolder_template_dir.resolve())

    recursive_render(
        template_dir=dotfolder_template_dir.resolve(),
        environment=env,
        _root_dir=example_project_dir.resolve(),
    )

    rendered_template = example_project_dir / dotfolder_template.name
    assert rendered_template.exists()

    assert set(example_project_dir.rglob("**/*")) == preexisting_paths.union(
        {example_project_dir / rendered_template}
    )



from semantic_release.hvcs._base import HvcsBase
from semantic_release.hvcs.bitbucket import Bitbucket
from semantic_release.hvcs.gitea import Gitea
from semantic_release.hvcs.github import Github
from semantic_release.hvcs.gitlab import Gitlab
from semantic_release.hvcs.remote_hvcs_base import RemoteHvcsBase
from semantic_release.hvcs.token_auth import TokenAuth

__all__ = [
    "Bitbucket",
    "Gitea",
    "Github",
    "Gitlab",
    "HvcsBase",
    "RemoteHvcsBase",
    "TokenAuth",
]



from __future__ import annotations

import os
from typing import TYPE_CHECKING
from unittest import mock

import gitlab
import gitlab.exceptions
import gitlab.mixins
import gitlab.v4.objects
import pytest

from semantic_release.hvcs.gitlab import Gitlab

from tests.const import (
    EXAMPLE_HVCS_DOMAIN,
    EXAMPLE_REPO_NAME,
    EXAMPLE_REPO_OWNER,
    RELEASE_NOTES,
)

if TYPE_CHECKING:
    from typing import Generator


# Note: there's nothing special about the value of these variables,
# they're just constants for easier consistency with the faked objects
A_GOOD_TAG = "v1.2.3"
A_BAD_TAG = "v2.1.1-rc.1"
A_LOCKED_TAG = "v0.9.0"
A_MISSING_TAG = "v1.0.0+missing"
# But note this is the only ref we're making a "fake" commit for, so
# tests which need to query the remote for "a" ref, the exact sha for
# which doesn't matter, all use this constant
REF = "hashashash"


@pytest.fixture
def default_gl_project(example_git_https_url: str):
    return gitlab.Gitlab(url=example_git_https_url).projects.get(
        f"{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}", lazy=True
    )


@pytest.fixture
def default_gl_client(
    example_git_https_url: str,
    default_gl_project: gitlab.v4.objects.Project,
) -> Generator[Gitlab, None, None]:
    gitlab_client = Gitlab(remote_url=example_git_https_url)

    # make sure that when project tries to get the project instance, we return the mock
    # that we control
    project_get_mock = mock.patch.object(
        gitlab_client._client.projects,
        gitlab_client._client.projects.get.__name__,
        return_value=default_gl_project,
    )

    env_mock = mock.patch.dict(os.environ, {}, clear=True)

    with project_get_mock, env_mock:
        yield gitlab_client


@pytest.mark.parametrize(
    "patched_os_environ, hvcs_domain, expected_hvcs_domain, insecure",
    # NOTE: GitLab does not have a different api domain
    [
        # Default values
        ({}, None, f"https://{Gitlab.DEFAULT_DOMAIN}", False),
        (
            # Gather domain from environment
            {"CI_SERVER_URL": "https://special.custom.server/"},
            None,
            "https://special.custom.server",
            False,
        ),
        (
            # Custom domain with path prefix (derives from environment)
            {"CI_SERVER_URL": "https://special.custom.server/vcs/"},
            None,
            "https://special.custom.server/vcs",
            False,
        ),
        (
            # Ignore environment & use provided parameter value (ie from user config)
            {
                "CI_SERVER_URL": "https://special.custom.server/",
                "CI_API_V4_URL": "https://special.custom.server/api/v3",
            },
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            False,
        ),
        (
            # Allow insecure http connections explicitly
            {},
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            True,
        ),
        (
            # Infer insecure connection from user configuration
            {},
            EXAMPLE_HVCS_DOMAIN,
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            True,
        ),
    ],
)
@pytest.mark.parametrize(
    "remote_url",
    [
        f"git@{Gitlab.DEFAULT_DOMAIN}:{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
        f"https://{Gitlab.DEFAULT_DOMAIN}/{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
    ],
)
@pytest.mark.parametrize("token", ("abc123", None))
def test_gitlab_client_init(
    patched_os_environ: dict[str, str],
    hvcs_domain: str | None,
    expected_hvcs_domain: str,
    remote_url: str,
    token: str | None,
    insecure: bool,
):
    with mock.patch.dict(os.environ, patched_os_environ, clear=True):
        client = Gitlab(
            remote_url=remote_url,
            hvcs_domain=hvcs_domain,
            token=token,
            allow_insecure=insecure,
        )

        # Evaluate (expected -> actual)
        assert expected_hvcs_domain == client.hvcs_domain.url
        assert token == client.token
        assert remote_url == client._remote_url


@pytest.mark.parametrize(
    "hvcs_domain, insecure",
    [
        (f"ftp://{EXAMPLE_HVCS_DOMAIN}", False),
        (f"ftp://{EXAMPLE_HVCS_DOMAIN}", True),
        (f"http://{EXAMPLE_HVCS_DOMAIN}", False),
    ],
)
def test_gitlab_client_init_with_invalid_scheme(
    hvcs_domain: str,
    insecure: bool,
):
    with pytest.raises(ValueError), mock.patch.dict(os.environ, {}, clear=True):
        Gitlab(
            remote_url=f"https://{EXAMPLE_HVCS_DOMAIN}/{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
            hvcs_domain=hvcs_domain,
            allow_insecure=insecure,
        )


@pytest.mark.parametrize(
    "patched_os_environ, expected_owner, expected_name",
    [
        ({}, None, None),
        (
            {"CI_PROJECT_NAMESPACE": "path/to/repo", "CI_PROJECT_NAME": "foo"},
            "path/to/repo",
            "foo",
        ),
    ],
)
def test_gitlab_get_repository_owner_and_name(
    default_gl_client: Gitlab,
    example_git_https_url: str,
    patched_os_environ: dict[str, str],
    expected_owner: str | None,
    expected_name: str | None,
):
    # expected results should be a tuple[namespace, repo_name] and if both are None,
    # then the default value from GitLab class should be used
    expected_result = (expected_owner, expected_name)
    if expected_owner is None and expected_name is None:
        expected_result = super(
            Gitlab, default_gl_client
        )._get_repository_owner_and_name()

    with mock.patch.dict(os.environ, patched_os_environ, clear=True):
        # Execute in mocked environment
        result = Gitlab(
            remote_url=example_git_https_url,
        )._get_repository_owner_and_name()

        # Evaluate (expected -> actual)
        assert expected_result == result


@pytest.mark.parametrize(
    "use_token, token, remote_url, expected_auth_url",
    [
        (
            False,
            "",
            f"git@{Gitlab.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Gitlab.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            True,
            "",
            f"git@{Gitlab.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Gitlab.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            False,
            "aabbcc",
            f"git@{Gitlab.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Gitlab.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            True,
            "aabbcc",
            f"git@{Gitlab.DEFAULT_DOMAIN}:custom/example.git",
            f"https://gitlab-ci-token:aabbcc@{Gitlab.DEFAULT_DOMAIN}/custom/example.git",
        ),
    ],
)
def test_remote_url(
    use_token: bool,
    token: str,
    remote_url: str,
    expected_auth_url: str,
):
    with mock.patch.dict(os.environ, {}, clear=True):
        gl_client = Gitlab(remote_url=remote_url, token=token)

    assert expected_auth_url == gl_client.remote_url(use_token=use_token)


def test_compare_url(default_gl_client: Gitlab):
    start_rev = "revA"
    end_rev = "revB"
    expected_url = "{server}/{owner}/{repo}/-/compare/{from_rev}...{to_rev}".format(
        server=default_gl_client.hvcs_domain.url,
        owner=default_gl_client.owner,
        repo=default_gl_client.repo_name,
        from_rev=start_rev,
        to_rev=end_rev,
    )
    actual_url = default_gl_client.compare_url(from_rev=start_rev, to_rev=end_rev)
    assert expected_url == actual_url


def test_commit_hash_url(default_gl_client: Gitlab):
    expected_url = "{server}/{owner}/{repo}/-/commit/{sha}".format(
        server=default_gl_client.hvcs_domain.url,
        owner=default_gl_client.owner,
        repo=default_gl_client.repo_name,
        sha=REF,
    )
    assert expected_url == default_gl_client.commit_hash_url(REF)


def test_commit_hash_url_w_custom_server():
    """
    Test the commit hash URL generation for a self-hosted Bitbucket server with prefix.

    ref: https://github.com/python-semantic-release/python-semantic-release/issues/1204
    """
    sha = "244f7e11bcb1e1ce097db61594056bc2a32189a0"
    expected_url = "{server}/{owner}/{repo}/-/commit/{sha}".format(
        server=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo",
        owner="foo",
        repo=EXAMPLE_REPO_NAME,
        sha=sha,
    )

    with mock.patch.dict(os.environ, {}, clear=True):
        actual_url = Gitlab(
            remote_url=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo/foo/{EXAMPLE_REPO_NAME}.git",
            hvcs_domain=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo",
        ).commit_hash_url(sha)

    assert expected_url == actual_url


@pytest.mark.parametrize("issue_number", (666, "666", "#666"))
def test_issue_url(default_gl_client: Gitlab, issue_number: int | str):
    expected_url = "{server}/{owner}/{repo}/-/issues/{issue_num}".format(
        server=default_gl_client.hvcs_domain.url,
        owner=default_gl_client.owner,
        repo=default_gl_client.repo_name,
        issue_num=str(issue_number).lstrip("#"),
    )
    actual_url = default_gl_client.issue_url(issue_num=issue_number)
    assert expected_url == actual_url


@pytest.mark.parametrize("pr_number", (666, "666", "!666"))
def test_pull_request_url(default_gl_client: Gitlab, pr_number: int | str):
    expected_url = "{server}/{owner}/{repo}/-/merge_requests/{pr_number}".format(
        server=default_gl_client.hvcs_domain.url,
        owner=default_gl_client.owner,
        repo=default_gl_client.repo_name,
        pr_number=str(pr_number).lstrip("!"),
    )
    actual_url = default_gl_client.pull_request_url(pr_number=pr_number)
    assert expected_url == actual_url


@pytest.mark.parametrize("tag", (A_GOOD_TAG, A_LOCKED_TAG))
def test_create_release_succeeds(
    default_gl_client: Gitlab, default_gl_project: gitlab.v4.objects.Project, tag: str
):
    with mock.patch.object(
        default_gl_project.releases,
        default_gl_project.releases.create.__name__,
    ) as mocked_create_release:
        result = default_gl_client.create_release(tag, RELEASE_NOTES)

        assert tag == result
        mocked_create_release.assert_called_once_with(
            {
                "name": tag,
                "tag_name": tag,
                "tag_message": tag,
                "description": RELEASE_NOTES,
            }
        )


def test_create_release_fails_with_bad_tag(
    default_gl_client: Gitlab,
    default_gl_project: gitlab.v4.objects.Project,
):
    bad_request = gitlab.GitlabCreateError("401 Unauthorized")
    mock_failed_create = mock.patch.object(
        default_gl_project.releases,
        default_gl_project.releases.create.__name__,
        side_effect=bad_request,
    )
    with mock_failed_create, pytest.raises(gitlab.GitlabCreateError):
        default_gl_client.create_release(A_BAD_TAG, RELEASE_NOTES)


@pytest.mark.parametrize("tag", (A_GOOD_TAG, A_LOCKED_TAG))
def test_update_release_succeeds(default_gl_client: Gitlab, tag: str):
    fake_release_obj = gitlab.v4.objects.ProjectReleaseManager(
        default_gl_client._client
    ).get(tag, lazy=True)
    fake_release_obj._attrs["name"] = tag

    with mock.patch.object(
        gitlab.mixins.SaveMixin,
        gitlab.mixins.SaveMixin.save.__name__,
    ) as mocked_update_release:
        release_id = default_gl_client.edit_release_notes(
            fake_release_obj, RELEASE_NOTES
        )

        assert tag == release_id
        mocked_update_release.assert_called_once()
        assert RELEASE_NOTES == fake_release_obj.description  # noqa: SIM300


def test_update_release_fails_with_missing_tag(
    default_gl_client: Gitlab,
    default_gl_project: gitlab.v4.objects.Project,
):
    fake_release_obj = gitlab.v4.objects.ProjectRelease(
        default_gl_project.manager,
        {"id": A_MISSING_TAG, "name": A_MISSING_TAG},
        lazy=True,
    )
    mocked_update_release = mock.patch.object(
        gitlab.mixins.SaveMixin,
        gitlab.mixins.SaveMixin.save.__name__,
        side_effect=gitlab.GitlabUpdateError,
    )

    with mocked_update_release, pytest.raises(gitlab.GitlabUpdateError):
        default_gl_client.edit_release_notes(fake_release_obj, RELEASE_NOTES)


@pytest.mark.parametrize("prerelease", (True, False))
def test_create_or_update_release_when_create_succeeds(
    default_gl_client: Gitlab, prerelease: bool
):
    with mock.patch.object(
        default_gl_client,
        default_gl_client.create_release.__name__,
        return_value=A_GOOD_TAG,
    ) as mock_create_release, mock.patch.object(
        default_gl_client,
        default_gl_client.edit_release_notes.__name__,
        return_value=A_GOOD_TAG,
    ) as mock_edit_release_notes:
        # Execute in mock environment
        result = default_gl_client.create_or_update_release(
            A_GOOD_TAG, RELEASE_NOTES, prerelease
        )

        # Evaluate (expected -> actual)
        assert A_GOOD_TAG == result  # noqa: SIM300
        mock_create_release.assert_called_once_with(
            tag=A_GOOD_TAG, release_notes=RELEASE_NOTES, prerelease=prerelease
        )
        mock_edit_release_notes.assert_not_called()


def test_get_release_id_by_tag(
    default_gl_client: Gitlab,
    default_gl_project: gitlab.v4.objects.Project,
):
    dummy_release = default_gl_project.releases.get(A_GOOD_TAG, lazy=True)

    with mock.patch.object(
        default_gl_project.releases,
        default_gl_project.releases.get.__name__,
        return_value=dummy_release,
    ) as mocked_get_release_id:
        result = default_gl_client.get_release_by_tag(A_GOOD_TAG)

        assert dummy_release == result
        mocked_get_release_id.assert_called_once_with(A_GOOD_TAG)


def test_get_release_id_by_tag_fails(
    default_gl_client: Gitlab,
    default_gl_project: gitlab.v4.objects.Project,
):
    mocked_get_release_id = mock.patch.object(
        default_gl_project.releases,
        default_gl_project.releases.get.__name__,
        side_effect=gitlab.exceptions.GitlabAuthenticationError,
    )

    with pytest.raises(
        gitlab.exceptions.GitlabAuthenticationError
    ), mocked_get_release_id:
        default_gl_client.get_release_by_tag(A_GOOD_TAG)


def test_get_release_id_by_tag_not_found(
    default_gl_client: Gitlab,
    default_gl_project: gitlab.v4.objects.Project,
):
    mocked_get_release_id = mock.patch.object(
        default_gl_project.releases,
        default_gl_project.releases.get.__name__,
        side_effect=gitlab.exceptions.GitlabGetError,
    )

    with mocked_get_release_id:
        result = default_gl_client.get_release_by_tag(A_GOOD_TAG)

    assert result is None


@pytest.mark.parametrize("prerelease", (True, False))
def test_create_or_update_release_when_create_fails_and_update_succeeds(
    default_gl_client: Gitlab,
    prerelease: bool,
):
    bad_request = gitlab.GitlabCreateError("400 Bad Request")
    expected_release_obj = gitlab.v4.objects.ProjectRelease(
        gitlab.v4.objects.ProjectManager(default_gl_client._client),
        {"commit": {"id": "1"}, "name": A_GOOD_TAG},
        lazy=True,
    )

    with mock.patch.object(
        default_gl_client,
        default_gl_client.create_release.__name__,
        side_effect=bad_request,
    ), mock.patch.object(
        default_gl_client,
        default_gl_client.get_release_by_tag.__name__,
        return_value=expected_release_obj,
    ), mock.patch.object(
        default_gl_client,
        default_gl_client.edit_release_notes.__name__,
        return_value=A_GOOD_TAG,
    ) as mock_edit_release_notes:
        # Execute in mock environment
        default_gl_client.create_or_update_release(
            A_GOOD_TAG, RELEASE_NOTES, prerelease
        )

        # Evaluate (expected -> actual)
        mock_edit_release_notes.assert_called_once_with(
            release=expected_release_obj, release_notes=RELEASE_NOTES
        )


@pytest.mark.parametrize("prerelease", (True, False))
def test_create_or_update_release_when_create_fails_and_update_fails(
    default_gl_client: Gitlab,
    prerelease: bool,
):
    bad_request = gitlab.GitlabCreateError("400 Bad Request")
    not_found = gitlab.GitlabUpdateError("404 Not Found")
    fake_release_obj = gitlab.v4.objects.ProjectRelease(
        gitlab.v4.objects.ProjectManager(default_gl_client._client),
        {"commit": {"id": "1"}, "name": A_GOOD_TAG},
        lazy=True,
    )

    create_release_patch = mock.patch.object(
        default_gl_client,
        default_gl_client.create_release.__name__,
        side_effect=bad_request,
    )
    edit_release_notes_patch = mock.patch.object(
        default_gl_client,
        default_gl_client.edit_release_notes.__name__,
        side_effect=not_found,
    )
    get_release_by_id_patch = mock.patch.object(
        default_gl_client,
        default_gl_client.get_release_by_tag.__name__,
        return_value=fake_release_obj,
    )

    # Execute in mocked environment expecting a GitlabUpdateError to be raised
    with create_release_patch, edit_release_notes_patch, get_release_by_id_patch:  # noqa: SIM117
        with pytest.raises(gitlab.GitlabUpdateError):
            default_gl_client.create_or_update_release(
                A_GOOD_TAG, RELEASE_NOTES, prerelease
            )



from __future__ import annotations

import os
from datetime import datetime
from textwrap import dedent
from typing import TYPE_CHECKING
from unittest import mock

import pytest
from git import Commit, Object, Repo

from semantic_release.changelog.context import ChangelogMode, make_changelog_context
from semantic_release.changelog.release_history import Release, ReleaseHistory
from semantic_release.changelog.template import environment
from semantic_release.commit_parser import ParsedCommit
from semantic_release.enums import LevelBump
from semantic_release.hvcs import Bitbucket, Gitea, Github, Gitlab
from semantic_release.version.translator import Version

from tests.const import EXAMPLE_CHANGELOG_MD_CONTENT

if TYPE_CHECKING:
    from pathlib import Path

    from git import Actor

    from tests.fixtures.git_repo import BuildRepoFn


@pytest.fixture
def changelog_tpl_github_context() -> str:
    """Returns an changelog template which uses all the GitHub configured filters"""
    return dedent(
        r"""
        # Changelog

        > Repository: {{ "[%s](%s)" | format(context.hvcs_type | capitalize, "/" | create_repo_url) }}

        ## v2.0.0

        {{ "[Change Summary](%s)" | format("v1.0.0" | compare_url("v2.0.0")) }}

        ### Bug Fixes

        - Fixed a minor bug {{ "([#%s](%s))" | format(22, 22 | pull_request_url) }}
        - **cli:** fix a problem {{ "([%s](%s))" | format("000000", "000000" | commit_hash_url) }}

        ### Resolved Issues

        - {{ "[#%s](%s)" | format(19, 19 | issue_url) }}
        """
    ).lstrip()


@pytest.fixture
def changelog_tpl_gitea_context() -> str:
    """Returns an changelog template which uses all the Gitea configured filters"""
    return dedent(
        r"""
        # Changelog

        > Repository: {{ "[%s](%s)" | format(context.hvcs_type | capitalize, "/" | create_repo_url) }}

        ## v2.0.0

        ### Bug Fixes

        - Fixed a minor bug {{ "([#%s](%s))" | format(22, 22 | pull_request_url) }}
        - **cli:** fix a problem {{ "([%s](%s))" | format("000000", "000000" | commit_hash_url) }}

        ### Resolved Issues

        - {{ "[#%s](%s)" | format(19, 19 | issue_url) }}
        """
    ).lstrip()


@pytest.fixture
def changelog_tpl_gitlab_context() -> str:
    """Returns an changelog template which uses all the GitLab configured filters"""
    return dedent(
        r"""
        # Changelog

        > Repository: {{ "[%s](%s)" | format(context.hvcs_type | capitalize, "/" | create_repo_url) }}

        ## v2.0.0

        {{ "[Change Summary](%s)" | format("v1.0.0" | compare_url("v2.0.0")) }}

        ### Bug Fixes

        - Fixed a minor bug {{ "([#%s](%s))" | format(22, 22 | merge_request_url) }}
        - Fixed a performance bug {{ "([#%s](%s))" | format(25, 25 | pull_request_url) }}
        - **cli:** fix a problem {{ "([%s](%s))" | format("000000", "000000" | commit_hash_url) }}

        ### Resolved Issues

        - {{ "[#%s](%s)" | format(19, 19 | issue_url) }}
        """
    ).lstrip()


@pytest.fixture
def changelog_tpl_bitbucket_context() -> str:
    """Returns an changelog template which uses all the BitBucket configured filters"""
    return dedent(
        r"""
        # Changelog

        > Repository: {{ "[%s](%s)" | format(context.hvcs_type | capitalize, "/" | create_repo_url) }}

        ## v2.0.0

        {{ "[Change Summary](%s)" | format("v1.0.0" | compare_url("v2.0.0")) }}

        ### Bug Fixes

        - Fixed a minor bug {{ "([#%s](%s))" | format(22, 22 | pull_request_url) }}
        - **cli:** fix a problem {{ "([%s](%s))" | format("000000", "000000" | commit_hash_url) }}
        """
    ).lstrip()


@pytest.fixture
def artificial_release_history(commit_author: Actor):
    version = Version.parse("1.0.0")

    commit_subject = "fix(cli): fix a problem"

    fix_commit = Commit(
        Repo("."),
        Object.NULL_HEX_SHA[:20].encode("utf-8"),
        message=commit_subject,
    )

    fix_commit_parsed = ParsedCommit(
        bump=LevelBump.PATCH,
        type="fix",
        scope="cli",
        descriptions=[commit_subject],
        breaking_descriptions=[],
        commit=fix_commit,
    )

    commit_subject = "feat(cli): add a new feature"

    feat_commit = Commit(
        Repo("."),
        Object.NULL_HEX_SHA[:20].encode("utf-8"),
        message=commit_subject,
    )

    feat_commit_parsed = ParsedCommit(
        bump=LevelBump.MINOR,
        type="feat",
        scope="cli",
        descriptions=[commit_subject],
        breaking_descriptions=[],
        commit=feat_commit,
    )

    return ReleaseHistory(
        unreleased={
            "feature": [feat_commit_parsed],
        },
        released={
            version: Release(
                tagger=commit_author,
                committer=commit_author,
                tagged_date=datetime.now(),
                elements={
                    "feature": [feat_commit_parsed],
                    "fix": [fix_commit_parsed],
                },
                version=version,
            )
        },
    )


def test_changelog_context_bitbucket(
    changelog_tpl_bitbucket_context: str,
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    hvcs = Bitbucket(example_git_https_url)

    expected_changelog = str.join(
        "\n",
        [
            "# Changelog",
            "",
            f'> Repository: [{hvcs.__class__.__name__.capitalize()}]({hvcs.create_repo_url("")})',
            "",
            "## v2.0.0",
            "",
            f'[Change Summary]({hvcs.compare_url("v1.0.0", "v2.0.0")})',
            "",
            "### Bug Fixes",
            "",
            f"- Fixed a minor bug ([#22]({hvcs.pull_request_url(22)}))",
            f'- **cli:** fix a problem ([000000]({hvcs.commit_hash_url("000000")}))',
            "",
        ],
    )

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=hvcs,
        release_history=artificial_release_history,
        mode=ChangelogMode.INIT,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl_bitbucket_context).render()

    # Evaluate
    assert expected_changelog == actual_changelog


def test_changelog_context_github(
    changelog_tpl_github_context: str,
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    hvcs = Github(example_git_https_url)

    expected_changelog = str.join(
        "\n",
        [
            "# Changelog",
            "",
            f'> Repository: [{hvcs.__class__.__name__.capitalize()}]({hvcs.create_repo_url("")})',
            "",
            "## v2.0.0",
            "",
            f'[Change Summary]({hvcs.compare_url("v1.0.0", "v2.0.0")})',
            "",
            "### Bug Fixes",
            "",
            f"- Fixed a minor bug ([#22]({hvcs.pull_request_url(22)}))",
            f'- **cli:** fix a problem ([000000]({hvcs.commit_hash_url("000000")}))',
            "",
            "### Resolved Issues",
            "",
            f"- [#19]({hvcs.issue_url(19)})",
            "",
        ],
    )

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=hvcs,
        release_history=artificial_release_history,
        mode=ChangelogMode.INIT,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl_github_context).render()

    # Evaluate
    assert expected_changelog == actual_changelog


def test_changelog_context_gitea(
    changelog_tpl_gitea_context: str,
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    hvcs = Gitea(example_git_https_url)

    expected_changelog = str.join(
        "\n",
        [
            "# Changelog",
            "",
            f'> Repository: [{hvcs.__class__.__name__.capitalize()}]({hvcs.create_repo_url("")})',
            "",
            "## v2.0.0",
            "",
            "### Bug Fixes",
            "",
            f"- Fixed a minor bug ([#22]({hvcs.pull_request_url(22)}))",
            f'- **cli:** fix a problem ([000000]({hvcs.commit_hash_url("000000")}))',
            "",
            "### Resolved Issues",
            "",
            f"- [#19]({hvcs.issue_url(19)})",
            "",
        ],
    )

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=hvcs,
        release_history=artificial_release_history,
        mode=ChangelogMode.INIT,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl_gitea_context).render()

    # Evaluate
    assert expected_changelog == actual_changelog


def test_changelog_context_gitlab(
    changelog_tpl_gitlab_context: str,
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    hvcs = Gitlab(example_git_https_url)

    expected_changelog = str.join(
        "\n",
        [
            "# Changelog",
            "",
            f'> Repository: [{hvcs.__class__.__name__.capitalize()}]({hvcs.create_repo_url("")})',
            "",
            "## v2.0.0",
            "",
            f'[Change Summary]({hvcs.compare_url("v1.0.0", "v2.0.0")})',
            "",
            "### Bug Fixes",
            "",
            f"- Fixed a minor bug ([#22]({hvcs.merge_request_url(22)}))",
            f"- Fixed a performance bug ([#25]({hvcs.pull_request_url(25)}))",
            f'- **cli:** fix a problem ([000000]({hvcs.commit_hash_url("000000")}))',
            "",
            "### Resolved Issues",
            "",
            f"- [#19]({hvcs.issue_url(19)})",
            "",
        ],
    )

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=hvcs,
        release_history=artificial_release_history,
        mode=ChangelogMode.INIT,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl_gitlab_context).render()

    # Evaluate
    assert expected_changelog == actual_changelog


def test_changelog_context_read_file(
    example_git_https_url: str,
    build_configured_base_repo: BuildRepoFn,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
    change_to_ex_proj_dir: Path,
    example_project_dir: Path,
):
    build_configured_base_repo(example_project_dir)

    # normalize expected to os specific newlines
    expected_changelog = str.join(
        os.linesep,
        [
            *[
                line.replace("\r", "")
                for line in EXAMPLE_CHANGELOG_MD_CONTENT.strip().split("\n")
            ],
            "",
        ],
    )

    changelog_tpl = """{{ "%s" | read_file | trim }}%ls""".replace(
        "%s", str(changelog_md_file)
    ).replace("%ls", os.linesep)

    env = environment(
        newline_sequence="\n" if os.linesep == "\n" else "\r\n",
        trim_blocks=True,
        lstrip_blocks=True,
        keep_trailing_newline=True,
        autoescape=False,
    )
    context = make_changelog_context(
        hvcs_client=Gitlab(example_git_https_url),
        release_history=artificial_release_history,
        mode=ChangelogMode.UPDATE,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl).render()

    # Evaluate
    assert expected_changelog.encode() == actual_changelog.encode()


@pytest.mark.parametrize("file_path", ["", "nonexistent.md"])
def test_changelog_context_read_file_fails_gracefully(
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
    file_path: str,
):
    changelog_tpl = """{{ "%s" | read_file }}""".replace("%s", file_path)
    expected_changelog = ""

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=Gitlab(example_git_https_url),
        release_history=artificial_release_history,
        mode=ChangelogMode.UPDATE,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl).render()

    # Evaluate
    assert expected_changelog == actual_changelog


def test_changelog_context_autofit_text_width(
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    changelog_tpl = """{{ "This is a long line that should be autofitted" | autofit_text_width(20) }}"""
    expected_changelog = "This is a long line\nthat should be\nautofitted"

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=Gitlab(example_git_https_url),
        release_history=artificial_release_history,
        mode=ChangelogMode.UPDATE,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl).render()

    # Evaluate
    assert expected_changelog == actual_changelog


def test_changelog_context_autofit_text_width_w_indent(
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    changelog_tpl = """{{ "This is a long line that should be autofitted" | autofit_text_width(20, indent_size=2) }}"""
    expected_changelog = "This is a long line\n  that should be\n  autofitted"

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=Gitlab(example_git_https_url),
        release_history=artificial_release_history,
        mode=ChangelogMode.UPDATE,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl).render()

    # Evaluate
    assert expected_changelog == actual_changelog


def test_changelog_context_sort_numerically(
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    changelog_tpl = dedent(
        """\
        {{ [
                ".. _#5: link",
                ".. _PR#3: link",
                ".. _PR#10: link",
                ".. _#100: link"
            ] | sort_numerically | join("\\n")
        }}
        """
    )

    expected_changelog = dedent(
        """\
        .. _#5: link
        .. _#100: link
        .. _PR#3: link
        .. _PR#10: link
        """
    )

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=Gitlab(example_git_https_url),
        release_history=artificial_release_history,
        mode=ChangelogMode.UPDATE,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl).render()

    # Evaluate
    assert expected_changelog == actual_changelog


def test_changelog_context_sort_numerically_reverse(
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    changelog_tpl = dedent(
        """\
        {{ [
                ".. _#5: link",
                ".. _PR#3: link",
                ".. _PR#10: link",
                ".. _#100: link"
            ] | sort_numerically(reverse=True) | join("\\n")
        }}
        """
    )

    expected_changelog = dedent(
        """\
        .. _#100: link
        .. _#5: link
        .. _PR#10: link
        .. _PR#3: link
        """
    )

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=Gitlab(example_git_https_url),
        release_history=artificial_release_history,
        mode=ChangelogMode.UPDATE,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl).render()

    # Evaluate
    assert expected_changelog == actual_changelog


def test_changelog_context_pypi_url_filter(
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    changelog_tpl = dedent(
        """\
        {{ "example-package" | create_pypi_url }}
        """
    )

    expected_changelog = dedent(
        """\
        https://pypi.org/project/example-package
        """
    )

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=Gitlab(example_git_https_url),
        release_history=artificial_release_history,
        mode=ChangelogMode.UPDATE,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl).render()

    # Evaluate
    assert expected_changelog == actual_changelog


def test_changelog_context_pypi_url_filter_tagged(
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    version = "1.0.0"
    changelog_tpl = dedent(
        """\
        {%    set release = context.history.released.values() | first
        %}{{
              "example-package" | create_pypi_url(release.version | string)
        }}
        """
    )

    expected_changelog = dedent(
        f"""\
        https://pypi.org/project/example-package/{version}
        """
    )

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=Gitlab(example_git_https_url),
        release_history=artificial_release_history,
        mode=ChangelogMode.UPDATE,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl).render()

    # Evaluate
    assert expected_changelog == actual_changelog


@pytest.mark.parametrize("hvcs_client_class", [Github, Gitlab, Gitea])
def test_changelog_context_release_url_filter(
    example_git_https_url: str,
    hvcs_client_class: type[Github | Gitlab | Gitea],
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    version = list(artificial_release_history.released.keys())[-1]

    changelog_tpl = dedent(
        """\
        {%      set release = context.history.released.values() | first
        %}{{
                "[%s](%s)" | format(
                    release.version.as_tag(),
                    release.version.as_tag() | create_release_url,
                )
        }}
        """
    )

    with mock.patch.dict(os.environ, {}, clear=True):
        hvcs_client = hvcs_client_class(remote_url=example_git_https_url)
        expected_changelog = dedent(
            f"""\
            [{version.as_tag()}]({hvcs_client.create_release_url(version.as_tag())})
            """
        )

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=hvcs_client,
        release_history=artificial_release_history,
        mode=ChangelogMode.UPDATE,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl).render()

    # Evaluate
    assert expected_changelog == actual_changelog


@pytest.mark.parametrize("hvcs_client_class", [Github, Gitlab, Gitea, Bitbucket])
def test_changelog_context_format_w_official_name_filter(
    example_git_https_url: str,
    hvcs_client_class: type[Github | Gitlab | Gitea],
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
):
    changelog_tpl = dedent(
        """\
        {{ "%s" | format_w_official_vcs_name }}
        {{ "{}" | format_w_official_vcs_name }}
        {{ "{vcs_name}" | format_w_official_vcs_name }}
        """
    )

    with mock.patch.dict(os.environ, {}, clear=True):
        hvcs_client = hvcs_client_class(remote_url=example_git_https_url)
        expected_changelog = dedent(
            f"""\
            {hvcs_client.OFFICIAL_NAME}
            {hvcs_client.OFFICIAL_NAME}
            {hvcs_client.OFFICIAL_NAME}
            """
        )

    env = environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
    context = make_changelog_context(
        hvcs_client=hvcs_client,
        release_history=artificial_release_history,
        mode=ChangelogMode.UPDATE,
        prev_changelog_file=changelog_md_file,
        insertion_flag="",
        mask_initial_release=False,
    )
    context.bind_to_environment(env)

    # Create changelog from template with environment
    actual_changelog = env.from_string(changelog_tpl).render()

    # Evaluate
    assert expected_changelog == actual_changelog



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

import pytest
from pytest_lazy_fixtures import lf as lazy_fixture

from tests.const import CHANGELOG_SUBCMD, EXAMPLE_PROJECT_LICENSE, MAIN_PROG_NAME
from tests.fixtures.repos import (
    repo_w_github_flow_w_default_release_channel_conventional_commits,
    repo_w_github_flow_w_feature_release_channel_conventional_commits,
    repo_w_trunk_only_conventional_commits,
    repo_w_trunk_only_emoji_commits,
    repo_w_trunk_only_scipy_commits,
)
from tests.util import assert_successful_exit_code

if TYPE_CHECKING:
    from requests_mock import Mocker

    from tests.conftest import GetCachedRepoDataFn, GetStableDateNowFn, RunCliFn
    from tests.fixtures.example_project import UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import (
        BuiltRepoResult,
        GenerateDefaultReleaseNotesFromDefFn,
        GetCfgValueFromDefFn,
        GetHvcsClientFromRepoDefFn,
        GetVersionsFromRepoBuildDefFn,
        SplitRepoActionsByReleaseTagsFn,
    )


@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_fixture_name)
        for repo_fixture_name in [
            repo_w_trunk_only_conventional_commits.__name__,
        ]
    ],
)
def test_changelog_latest_release_notes(
    repo_result: BuiltRepoResult,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    get_hvcs_client_from_repo_def: GetHvcsClientFromRepoDefFn,
    run_cli: RunCliFn,
    post_mocker: Mocker,
    split_repo_actions_by_release_tags: SplitRepoActionsByReleaseTagsFn,
    generate_default_release_notes_from_def: GenerateDefaultReleaseNotesFromDefFn,
):
    # Setup
    repo_def = repo_result["definition"]
    tag_format_str: str = get_cfg_value_from_def(repo_def, "tag_format_str")  # type: ignore[assignment]
    repo_actions_per_version = split_repo_actions_by_release_tags(
        repo_definition=repo_def
    )
    all_versions = get_versions_from_repo_build_def(repo_def)
    latest_release_version = all_versions[-1]
    release_tag = tag_format_str.format(version=latest_release_version)

    expected_release_notes = generate_default_release_notes_from_def(
        version_actions=repo_actions_per_version[latest_release_version],
        hvcs=get_hvcs_client_from_repo_def(repo_def),
        previous_version=(all_versions[-2] if len(all_versions) > 1 else None),
        license_name=EXAMPLE_PROJECT_LICENSE,
        mask_initial_release=get_cfg_value_from_def(repo_def, "mask_initial_release"),
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD, "--post-to-release-tag", release_tag]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert post_mocker.call_count == 1
    assert post_mocker.last_request is not None
    request_body = post_mocker.last_request.json()

    assert "body" in request_body
    actual_posted_notes = request_body["body"]

    assert expected_release_notes == actual_posted_notes


@pytest.mark.parametrize(
    "repo_result, mask_initial_release",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            True,
        ),
        pytest.param(
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            False,
            marks=pytest.mark.comprehensive,
        ),
        *[
            pytest.param(
                lazy_fixture(repo_fixture_name),
                mask_initial_release,
                marks=pytest.mark.comprehensive,
            )
            for mask_initial_release in [True, False]
            for repo_fixture_name in [
                repo_w_github_flow_w_default_release_channel_conventional_commits.__name__,
            ]
        ],
    ],
)
def test_changelog_previous_release_notes(
    repo_result: BuiltRepoResult,
    mask_initial_release: bool,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    get_hvcs_client_from_repo_def: GetHvcsClientFromRepoDefFn,
    run_cli: RunCliFn,
    post_mocker: Mocker,
    split_repo_actions_by_release_tags: SplitRepoActionsByReleaseTagsFn,
    generate_default_release_notes_from_def: GenerateDefaultReleaseNotesFromDefFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
):
    # Setup
    repo_def = repo_result["definition"]
    tag_format_str: str = get_cfg_value_from_def(repo_def, "tag_format_str")  # type: ignore[assignment]
    repo_actions_per_version = split_repo_actions_by_release_tags(
        repo_definition=repo_def
    )
    # Extract all versions except for the latest one
    all_prev_versions = get_versions_from_repo_build_def(repo_def)[:-1]
    latest_release_version = all_prev_versions[-1]
    release_tag = tag_format_str.format(version=latest_release_version)

    expected_release_notes = generate_default_release_notes_from_def(
        version_actions=repo_actions_per_version[latest_release_version],
        hvcs=get_hvcs_client_from_repo_def(repo_def),
        previous_version=(
            all_prev_versions[-2] if len(all_prev_versions) > 1 else None
        ),
        license_name=EXAMPLE_PROJECT_LICENSE,
        mask_initial_release=mask_initial_release,
    )

    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.mask_initial_release",
        mask_initial_release,
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD, "--post-to-release-tag", release_tag]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert post_mocker.call_count == 1
    assert post_mocker.last_request is not None
    request_body = post_mocker.last_request.json()

    assert "body" in request_body
    actual_posted_notes = request_body["body"]

    assert expected_release_notes == actual_posted_notes


@pytest.mark.parametrize(
    "repo_result, repo_fixture_name, mask_initial_release, license_name",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            repo_w_trunk_only_conventional_commits.__name__,
            True,
            "BSD-3-Clause",
        ),
        pytest.param(
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            repo_w_trunk_only_conventional_commits.__name__,
            False,
            "BSD-3-Clause",
            marks=pytest.mark.comprehensive,
        ),
        *[
            pytest.param(
                lazy_fixture(repo_fixture_name),
                repo_fixture_name,
                mask_initial_release,
                "BSD-3-Clause",
                marks=pytest.mark.comprehensive,
            )
            for mask_initial_release in [True, False]
            for repo_fixture_name in [
                repo_w_trunk_only_emoji_commits.__name__,
                repo_w_trunk_only_scipy_commits.__name__,
                # Add more repos here if needed
                # github_flow had issues as its hard to generate the release notes from squash commits
                repo_w_github_flow_w_feature_release_channel_conventional_commits.__name__,
            ]
        ],
    ],
)
def test_changelog_release_notes_license_change(
    repo_result: BuiltRepoResult,
    license_name: str,
    mask_initial_release: bool,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    get_hvcs_client_from_repo_def: GetHvcsClientFromRepoDefFn,
    run_cli: RunCliFn,
    post_mocker: Mocker,
    split_repo_actions_by_release_tags: SplitRepoActionsByReleaseTagsFn,
    generate_default_release_notes_from_def: GenerateDefaultReleaseNotesFromDefFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    repo_fixture_name: str,
    stable_now_date: GetStableDateNowFn,
    get_cached_repo_data: GetCachedRepoDataFn,
):
    # Setup
    repo_def = repo_result["definition"]
    tag_format_str: str = get_cfg_value_from_def(repo_def, "tag_format_str")  # type: ignore[assignment]

    if not (repo_build_data := get_cached_repo_data(repo_fixture_name)):
        pytest.fail("Repo build date not found in cache")

    repo_build_datetime = datetime.strptime(repo_build_data["build_date"], "%Y-%m-%d")
    now_datetime = stable_now_date().replace(
        year=repo_build_datetime.year,
        month=repo_build_datetime.month,
        day=repo_build_datetime.day,
    )

    repo_actions_per_version = split_repo_actions_by_release_tags(
        repo_definition=repo_def,
    )
    # Extract all versions
    all_versions = get_versions_from_repo_build_def(repo_def)
    assert len(all_versions) > 1
    latest_release_version = all_versions[-1]
    previous_release_version = all_versions[-2]
    latest_release_tag = tag_format_str.format(version=latest_release_version)
    prev_release_tag = tag_format_str.format(version=previous_release_version)

    expected_release_notes = generate_default_release_notes_from_def(
        version_actions=repo_actions_per_version[latest_release_version],
        hvcs=get_hvcs_client_from_repo_def(repo_def),
        previous_version=(previous_release_version if len(all_versions) > 1 else None),
        license_name=license_name,
        mask_initial_release=mask_initial_release,
    )

    expected_prev_release_notes = generate_default_release_notes_from_def(
        version_actions=repo_actions_per_version[previous_release_version],
        hvcs=get_hvcs_client_from_repo_def(repo_def),
        previous_version=(all_versions[-3] if len(all_versions) > 2 else None),
        license_name=EXAMPLE_PROJECT_LICENSE,
        mask_initial_release=mask_initial_release,
    )

    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.mask_initial_release",
        mask_initial_release,
    )
    update_pyproject_toml("project.license-expression", license_name)

    git_repo = repo_result["repo"]

    git_repo.git.commit(
        amend=True,
        a=True,
        no_edit=True,
        date=now_datetime.isoformat(timespec="seconds"),
    )

    with git_repo.git.custom_environment(
        GIT_COMMITTER_DATE=now_datetime.isoformat(timespec="seconds"),
    ):
        git_repo.git.tag(latest_release_tag, d=True)
        git_repo.git.tag(latest_release_tag, a=True, m=latest_release_tag)

    # Act
    cli_cmd = [
        MAIN_PROG_NAME,
        CHANGELOG_SUBCMD,
        "--post-to-release-tag",
        latest_release_tag,
    ]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert post_mocker.call_count == 1
    assert post_mocker.last_request is not None
    request_body = post_mocker.last_request.json()

    assert "body" in request_body
    actual_new_posted_notes = request_body["body"]

    assert expected_release_notes == actual_new_posted_notes

    # Generate the previous release notes
    cli_cmd = [
        MAIN_PROG_NAME,
        CHANGELOG_SUBCMD,
        "--post-to-release-tag",
        prev_release_tag,
    ]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert post_mocker.call_count == 2
    assert post_mocker.last_request is not None
    request_body = post_mocker.last_request.json()

    assert "body" in request_body
    actual_prev_posted_notes = request_body["body"]

    assert expected_prev_release_notes == actual_prev_posted_notes

    assert actual_prev_posted_notes != actual_new_posted_notes



# ruff: noqa: SIM300
from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING, Iterable, Sequence

import pytest

from semantic_release.commit_parser.conventional import (
    ConventionalCommitParser,
    ConventionalCommitParserOptions,
)
from semantic_release.commit_parser.token import ParsedCommit, ParseError
from semantic_release.enums import LevelBump

from tests.const import SUPPORTED_ISSUE_CLOSURE_PREFIXES

if TYPE_CHECKING:
    from tests.conftest import MakeCommitObjFn


# NOTE: GitLab squash commits are not tested because by default
# they don't have any unique attributes of them and they are also
# fully customizable.
# See https://docs.gitlab.com/ee/user/project/merge_requests/commit_templates.html
# It also depends if Fast-Forward merge is enabled because that will
# define if there is a merge commit or not and with that likely no
# Merge Request Number included unless the user adds it.
# TODO: add the recommendation in the PSR documentation is to set your GitLab templates
# to mirror GitHub like references in the first subject line. Will Not matter
# if fast-forward merge is enabled or not.


@pytest.mark.parametrize(
    "commit_message", ["", "feat(parser\n): Add new parser pattern"]
)
def test_parser_raises_unknown_message_style(
    default_conventional_parser: ConventionalCommitParser,
    make_commit_obj: MakeCommitObjFn,
    commit_message: str,
):
    parsed_results = default_conventional_parser.parse(make_commit_obj(commit_message))
    assert isinstance(parsed_results, Iterable)
    for result in parsed_results:
        assert isinstance(result, ParseError)


@pytest.mark.parametrize(
    "commit_message, expected_commit_details",
    [
        pytest.param(
            commit_message,
            expected_commit_details,
            id=test_id,
        )
        for test_id, commit_message, expected_commit_details in [
            (
                "Single commit squashed via BitBucket PR resolution",
                dedent(
                    """\
                    Merged in feat/my-awesome-stuff  (pull request #10)

                    fix(release-config): some commit subject

                    An additional description

                    Second paragraph with multiple lines
                    that will be condensed

                    Resolves: #12
                    Signed-off-by: author <author@not-an-email.com>
                    """
                ),
                [
                    None,
                    {
                        "bump": LevelBump.PATCH,
                        "type": "bug fixes",
                        "scope": "release-config",
                        "descriptions": [
                            "some commit subject",
                            "An additional description",
                            "Second paragraph with multiple lines that will be condensed",
                            "Signed-off-by: author <author@not-an-email.com>",
                        ],
                        "linked_issues": ("#12",),
                        "linked_merge_request": "#10",
                    },
                ],
            ),
            (
                "Multiple commits squashed via BitBucket PR resolution",
                dedent(
                    """\
                    Merged in feat/my-awesome-stuff  (pull request #10)

                    fix(release-config): some commit subject

                    An additional description

                    Second paragraph with multiple lines
                    that will be condensed

                    Resolves: #12
                    Signed-off-by: author <author@not-an-email.com>

                    feat: implemented searching gizmos by keyword

                    docs(parser): add new parser pattern

                    fix(cli)!: changed option name

                    BREAKING CHANGE: A breaking change description

                    Closes: #555

                    invalid non-conventional formatted commit
                    """
                ),
                [
                    None,
                    {
                        "bump": LevelBump.PATCH,
                        "type": "bug fixes",
                        "scope": "release-config",
                        "descriptions": [
                            "some commit subject",
                            "An additional description",
                            "Second paragraph with multiple lines that will be condensed",
                            "Signed-off-by: author <author@not-an-email.com>",
                        ],
                        "linked_issues": ("#12",),
                        "linked_merge_request": "#10",
                    },
                    {
                        "bump": LevelBump.MINOR,
                        "type": "features",
                        "descriptions": ["implemented searching gizmos by keyword"],
                        "linked_merge_request": "#10",
                    },
                    {
                        "bump": LevelBump.NO_RELEASE,
                        "type": "documentation",
                        "scope": "parser",
                        "descriptions": [
                            "add new parser pattern",
                        ],
                        "linked_merge_request": "#10",
                    },
                    {
                        "bump": LevelBump.MAJOR,
                        "type": "bug fixes",
                        "scope": "cli",
                        "descriptions": [
                            "changed option name",
                            # This is a bit unusual but its because there is no identifier that will
                            # identify this as a separate commit so it gets included in the previous commit
                            "invalid non-conventional formatted commit",
                        ],
                        "breaking_descriptions": [
                            "A breaking change description",
                        ],
                        "linked_issues": ("#555",),
                        "linked_merge_request": "#10",
                    },
                ],
            ),
        ]
    ],
)
def test_parser_squashed_commit_bitbucket_squash_style(
    default_conventional_parser: ConventionalCommitParser,
    make_commit_obj: MakeCommitObjFn,
    commit_message: str,
    expected_commit_details: Sequence[dict | None],
):
    # Setup: Enable squash commit parsing
    parser = ConventionalCommitParser(
        options=ConventionalCommitParserOptions(
            **{
                **default_conventional_parser.options.__dict__,
                "parse_squash_commits": True,
            }
        )
    )

    # Build the commit object and parse it
    the_commit = make_commit_obj(commit_message)
    parsed_results = parser.parse(the_commit)

    # Validate the results
    assert isinstance(parsed_results, Iterable)
    assert (
        len(expected_commit_details) == len(parsed_results)
    ), f"Expected {len(expected_commit_details)} parsed results, but got {len(parsed_results)}"

    for result, expected in zip(parsed_results, expected_commit_details):
        if expected is None:
            assert isinstance(result, ParseError)
            continue

        assert isinstance(result, ParsedCommit)
        # Required
        assert expected["bump"] == result.bump
        assert expected["type"] == result.type
        # Optional
        assert expected.get("scope", "") == result.scope
        # TODO: v11 change to tuples
        assert expected.get("descriptions", []) == result.descriptions
        assert expected.get("breaking_descriptions", []) == result.breaking_descriptions
        assert expected.get("linked_issues", ()) == result.linked_issues
        assert expected.get("linked_merge_request", "") == result.linked_merge_request


@pytest.mark.parametrize(
    "commit_message, expected_commit_details",
    [
        pytest.param(
            commit_message,
            expected_commit_details,
            id=test_id,
        )
        for test_id, commit_message, expected_commit_details in [
            (
                "Single commit squashed via manual Git squash merge",
                dedent(
                    """\
                    Squashed commit of the following:

                    commit 63ec09b9e844e616dcaa7bae35a0b66671b59fbb
                    Author: author <author@not-an-email.com>
                    Date:   Sun Jan 19 12:05:23 2025 +0000

                        fix(release-config): some commit subject

                        An additional description

                        Second paragraph with multiple lines
                        that will be condensed

                        Resolves: #12
                        Signed-off-by: author <author@not-an-email.com>

                    """
                ),
                [
                    {
                        "bump": LevelBump.PATCH,
                        "type": "bug fixes",
                        "scope": "release-config",
                        "descriptions": [
                            "some commit subject",
                            "An additional description",
                            "Second paragraph with multiple lines that will be condensed",
                            "Signed-off-by: author <author@not-an-email.com>",
                        ],
                        "linked_issues": ("#12",),
                    }
                ],
            ),
            (
                "Multiple commits squashed via manual Git squash merge",
                dedent(
                    """\
                    Squashed commit of the following:

                    commit 63ec09b9e844e616dcaa7bae35a0b66671b59fbb
                    Author: author <author@not-an-email.com>
                    Date:   Sun Jan 19 12:05:23 2025 +0000

                        fix(release-config): some commit subject

                        An additional description

                        Second paragraph with multiple lines
                        that will be condensed

                        Resolves: #12
                        Signed-off-by: author <author@not-an-email.com>

                    commit 1f34769bf8352131ad6f4879b8c47becf3c7aa69
                    Author: author <author@not-an-email.com>
                    Date:   Sat Jan 18 10:13:53 2025 +0000

                        feat: implemented searching gizmos by keyword

                    commit b2334a64a11ef745a17a2a4034f651e08e8c45a6
                    Author: author <author@not-an-email.com>
                    Date:   Sat Jan 18 10:13:53 2025 +0000

                        docs(parser): add new parser pattern

                    commit 5f0292fb5a88c3a46e4a02bec35b85f5228e8e51
                    Author: author <author@not-an-email.com>
                    Date:   Sat Jan 18 10:13:53 2025 +0000

                        fix(cli)!: changed option name

                        BREAKING CHANGE: A breaking change description

                        Closes: #555

                    commit 2f314e7924be161cfbf220d3b6e2a6189a3b5609
                    Author: author <author@not-an-email.com>
                    Date:   Sat Jan 18 10:13:53 2025 +0000

                        invalid non-conventional formatted commit
                    """
                ),
                [
                    {
                        "bump": LevelBump.PATCH,
                        "type": "bug fixes",
                        "scope": "release-config",
                        "descriptions": [
                            "some commit subject",
                            "An additional description",
                            "Second paragraph with multiple lines that will be condensed",
                            "Signed-off-by: author <author@not-an-email.com>",
                        ],
                        "linked_issues": ("#12",),
                    },
                    {
                        "bump": LevelBump.MINOR,
                        "type": "features",
                        "descriptions": ["implemented searching gizmos by keyword"],
                    },
                    {
                        "bump": LevelBump.NO_RELEASE,
                        "type": "documentation",
                        "scope": "parser",
                        "descriptions": [
                            "add new parser pattern",
                        ],
                    },
                    {
                        "bump": LevelBump.MAJOR,
                        "type": "bug fixes",
                        "scope": "cli",
                        "descriptions": [
                            "changed option name",
                        ],
                        "breaking_descriptions": [
                            "A breaking change description",
                        ],
                        "linked_issues": ("#555",),
                    },
                    None,
                ],
            ),
        ]
    ],
)
def test_parser_squashed_commit_git_squash_style(
    default_conventional_parser: ConventionalCommitParser,
    make_commit_obj: MakeCommitObjFn,
    commit_message: str,
    expected_commit_details: Sequence[dict | None],
):
    # Setup: Enable squash commit parsing
    parser = ConventionalCommitParser(
        options=ConventionalCommitParserOptions(
            **{
                **default_conventional_parser.options.__dict__,
                "parse_squash_commits": True,
            }
        )
    )

    # Build the commit object and parse it
    the_commit = make_commit_obj(commit_message)
    parsed_results = parser.parse(the_commit)

    # Validate the results
    assert isinstance(parsed_results, Iterable)
    assert (
        len(expected_commit_details) == len(parsed_results)
    ), f"Expected {len(expected_commit_details)} parsed results, but got {len(parsed_results)}"

    for result, expected in zip(parsed_results, expected_commit_details):
        if expected is None:
            assert isinstance(result, ParseError)
            continue

        assert isinstance(result, ParsedCommit)
        # Required
        assert expected["bump"] == result.bump
        assert expected["type"] == result.type
        # Optional
        assert expected.get("scope", "") == result.scope
        # TODO: v11 change to tuples
        assert expected.get("descriptions", []) == result.descriptions
        assert expected.get("breaking_descriptions", []) == result.breaking_descriptions
        assert expected.get("linked_issues", ()) == result.linked_issues
        assert expected.get("linked_merge_request", "") == result.linked_merge_request


@pytest.mark.parametrize(
    "commit_message, expected_commit_details",
    [
        pytest.param(
            commit_message,
            expected_commit_details,
            id=test_id,
        )
        for test_id, commit_message, expected_commit_details in [
            (
                "Single commit squashed via GitHub PR resolution",
                dedent(
                    """\
                    fix(release-config): some commit subject (#10)

                    An additional description

                    Second paragraph with multiple lines
                    that will be condensed

                    Resolves: #12
                    Signed-off-by: author <author@not-an-email.com>
                    """
                ),
                [
                    {
                        "bump": LevelBump.PATCH,
                        "type": "bug fixes",
                        "scope": "release-config",
                        "descriptions": [
                            "some commit subject",
                            "An additional description",
                            "Second paragraph with multiple lines that will be condensed",
                            "Signed-off-by: author <author@not-an-email.com>",
                        ],
                        "linked_issues": ("#12",),
                        "linked_merge_request": "#10",
                    },
                ],
            ),
            (
                "Multiple commits squashed via GitHub PR resolution",
                dedent(
                    """\
                    fix(release-config): some commit subject (#10)

                    An additional description

                    Second paragraph with multiple lines
                    that will be condensed

                    Resolves: #12
                    Signed-off-by: author <author@not-an-email.com>

                    * feat: implemented searching gizmos by keyword

                    * docs(parser): add new parser pattern

                    * fix(cli)!: changed option name

                    BREAKING CHANGE: A breaking change description

                    Closes: #555

                    * invalid non-conventional formatted commit
                    """
                ),
                [
                    {
                        "bump": LevelBump.PATCH,
                        "type": "bug fixes",
                        "scope": "release-config",
                        "descriptions": [
                            "some commit subject",
                            "An additional description",
                            "Second paragraph with multiple lines that will be condensed",
                            "Signed-off-by: author <author@not-an-email.com>",
                        ],
                        "linked_issues": ("#12",),
                        "linked_merge_request": "#10",
                    },
                    {
                        "bump": LevelBump.MINOR,
                        "type": "features",
                        "descriptions": ["implemented searching gizmos by keyword"],
                        "linked_merge_request": "#10",
                    },
                    {
                        "bump": LevelBump.NO_RELEASE,
                        "type": "documentation",
                        "scope": "parser",
                        "descriptions": [
                            "add new parser pattern",
                        ],
                        "linked_merge_request": "#10",
                    },
                    {
                        "bump": LevelBump.MAJOR,
                        "type": "bug fixes",
                        "scope": "cli",
                        "descriptions": [
                            "changed option name",
                            # This is a bit unusual but its because there is no identifier that will
                            # identify this as a separate commit so it gets included in the previous commit
                            "* invalid non-conventional formatted commit",
                        ],
                        "breaking_descriptions": [
                            "A breaking change description",
                        ],
                        "linked_issues": ("#555",),
                        "linked_merge_request": "#10",
                    },
                ],
            ),
        ]
    ],
)
def test_parser_squashed_commit_github_squash_style(
    default_conventional_parser: ConventionalCommitParser,
    make_commit_obj: MakeCommitObjFn,
    commit_message: str,
    expected_commit_details: Sequence[dict | None],
):
    # Setup: Enable squash commit parsing
    parser = ConventionalCommitParser(
        options=ConventionalCommitParserOptions(
            **{
                **default_conventional_parser.options.__dict__,
                "parse_squash_commits": True,
            }
        )
    )

    # Build the commit object and parse it
    the_commit = make_commit_obj(commit_message)
    parsed_results = parser.parse(the_commit)

    # Validate the results
    assert isinstance(parsed_results, Iterable)
    assert (
        len(expected_commit_details) == len(parsed_results)
    ), f"Expected {len(expected_commit_details)} parsed results, but got {len(parsed_results)}"

    for result, expected in zip(parsed_results, expected_commit_details):
        if expected is None:
            assert isinstance(result, ParseError)
            continue

        assert isinstance(result, ParsedCommit)
        # Required
        assert expected["bump"] == result.bump
        assert expected["type"] == result.type
        # Optional
        assert expected.get("scope", "") == result.scope
        # TODO: v11 change to tuples
        assert expected.get("descriptions", []) == result.descriptions
        assert expected.get("breaking_descriptions", []) == result.breaking_descriptions
        assert expected.get("linked_issues", ()) == result.linked_issues
        assert expected.get("linked_merge_request", "") == result.linked_merge_request


@pytest.mark.parametrize(
    "commit_message, bump",
    [
        (
            "feat(parsers): add new parser pattern\n\nBREAKING CHANGE: change",
            LevelBump.MAJOR,
        ),
        ("feat(parsers)!: add new parser pattern", LevelBump.MAJOR),
        (
            "feat(parsers): add new parser pattern\n\nNew pattern is awesome\n\n"
            "BREAKING CHANGE: change \n",
            LevelBump.MAJOR,
        ),
        (
            "feat(parsers): add new parser pattern\n\nBREAKING-CHANGE: change !",
            LevelBump.MAJOR,
        ),
        ("feat(parser): add emoji parser", LevelBump.MINOR),
        ("fix(parser): fix regex in conventional parser", LevelBump.PATCH),
        ("test(parser): add a test for conventional parser", LevelBump.NO_RELEASE),
        ("feat(parser)!: edit data parsing stuff", LevelBump.MAJOR),
        ("fix!: edit data parsing stuff again", LevelBump.MAJOR),
        ("fix: superfix", LevelBump.PATCH),
    ],
)
def test_parser_returns_correct_bump_level(
    default_conventional_parser: ConventionalCommitParser,
    commit_message: str,
    bump: LevelBump,
    make_commit_obj: MakeCommitObjFn,
):
    parsed_results = default_conventional_parser.parse(make_commit_obj(commit_message))

    assert isinstance(parsed_results, Iterable)
    assert len(parsed_results) == 1

    result = next(iter(parsed_results))
    assert isinstance(result, ParsedCommit)
    assert result.bump is bump


@pytest.mark.parametrize(
    "message, type_",
    [
        ("feat(parser): ...", "features"),
        ("fix(parser): ...", "bug fixes"),
        ("test(parser): ...", "testing"),
        ("docs(parser): ...", "documentation"),
        ("style(parser): ...", "code style"),
        ("refactor(parser): ...", "refactoring"),
        ("chore(parser): ...", "chores"),
    ],
)
def test_parser_return_type_from_commit_message(
    default_conventional_parser: ConventionalCommitParser,
    message: str,
    type_: str,
    make_commit_obj: MakeCommitObjFn,
):
    parsed_results = default_conventional_parser.parse(make_commit_obj(message))

    assert isinstance(parsed_results, Iterable)
    assert len(parsed_results) == 1

    result = next(iter(parsed_results))
    assert isinstance(result, ParsedCommit)
    assert result.type == type_


@pytest.mark.parametrize(
    "message, scope",
    [
        ("chore(parser): ...", "parser"),
        ("chore(a part): ...", "a part"),
        ("chore(a_part): ...", "a_part"),
        ("chore(a-part): ...", "a-part"),
        ("chore(a.part): ...", "a.part"),
        ("chore(a+part): ...", "a+part"),
        ("chore(a&part): ...", "a&part"),
        ("chore((part)): ...", "(part)"),
        ("chore((p):rt): ...", "(p):rt"),
    ],
)
def test_parser_return_scope_from_commit_message(
    default_conventional_parser: ConventionalCommitParser,
    message: str,
    scope: str,
    make_commit_obj: MakeCommitObjFn,
):
    parsed_results = default_conventional_parser.parse(make_commit_obj(message))

    assert isinstance(parsed_results, Iterable)
    assert len(parsed_results) == 1

    result = next(iter(parsed_results))
    assert isinstance(result, ParsedCommit)
    assert result.scope == scope


_long_text = (
    "This is an long explanatory part of a commit message. It should give "
    "some insight to the fix this commit adds to the codebase."
)
_footer = "Closes: #400"


@pytest.mark.parametrize(
    "message, descriptions",
    [
        ("feat(parser): add emoji parser", ["add emoji parser"]),
        (
            "fix(parser): fix regex in conventional parser",
            ["fix regex in conventional parser"],
        ),
        (
            "test(parser): add a test for conventional parser",
            ["add a test for conventional parser"],
        ),
        (
            f"fix(tox): fix env \n\n{_long_text}\n\n{_footer}",
            ["fix env ", _long_text],
        ),
        ("fix: superfix", ["superfix"]),
    ],
)
def test_parser_return_subject_from_commit_message(
    default_conventional_parser: ConventionalCommitParser,
    message: str,
    descriptions: list[str],
    make_commit_obj: MakeCommitObjFn,
):
    parsed_results = default_conventional_parser.parse(make_commit_obj(message))

    assert isinstance(parsed_results, Iterable)
    assert len(parsed_results) == 1

    result = next(iter(parsed_results))
    assert isinstance(result, ParsedCommit)
    assert result.descriptions == descriptions


@pytest.mark.parametrize(
    "message, subject, merge_request_number",
    [
        # GitHub, Gitea style
        (
            "feat(parser): add emoji parser (#123)",
            "add emoji parser",
            "#123",
        ),
        # GitLab style
        (
            "fix(parser): fix regex in conventional parser (!456)",
            "fix regex in conventional parser",
            "!456",
        ),
        # BitBucket style
        (
            "feat(parser): add emoji parser (pull request #123)",
            "add emoji parser",
            "#123",
        ),
        # Both a linked merge request and an issue footer (should return the linked merge request)
        ("fix: superfix (#123)\n\nCloses: #400", "superfix", "#123"),
        # None
        ("fix: superfix", "superfix", ""),
        # None but includes an issue footer it should not be considered a linked merge request
        ("fix: superfix\n\nCloses: #400", "superfix", ""),
    ],
)
def test_parser_return_linked_merge_request_from_commit_message(
    default_conventional_parser: ConventionalCommitParser,
    message: str,
    subject: str,
    merge_request_number: str,
    make_commit_obj: MakeCommitObjFn,
):
    parsed_results = default_conventional_parser.parse(make_commit_obj(message))

    assert isinstance(parsed_results, Iterable)
    assert len(parsed_results) == 1

    result = next(iter(parsed_results))
    assert isinstance(result, ParsedCommit)
    assert merge_request_number == result.linked_merge_request
    assert subject == result.descriptions[0]


@pytest.mark.parametrize(
    "message, linked_issues",
    [
        *[
            # GitHub, Gitea, GitLab style
            (
                f"feat(parser): add magic parser\n\n{footer}",
                linked_issues,
            )
            for footer_prefix in SUPPORTED_ISSUE_CLOSURE_PREFIXES
            for footer, linked_issues in [
                # Single issue
                (
                    f"{footer_prefix.capitalize()}: #555",
                    ["#555"],
                ),  # Git Footer style (capitalized)
                (f"{footer_prefix.lower()}: #555", ["#555"]),  # lowercase prefix
                (f"{footer_prefix.upper()}: #555", ["#555"]),  # uppercase prefix
                # Mulitple issues (variant 1: list with one prefix, not supported by GitHub)
                (
                    f"{footer_prefix}: #555,#444",
                    ["#444", "#555"],
                ),  # Comma separated (w/o space)
                (
                    f"{footer_prefix}: #555, #444",
                    ["#444", "#555"],
                ),  # Comma separated (w/ space)
                (
                    f"{footer_prefix}: #555 , #444",
                    ["#444", "#555"],
                ),  # Comma separated (w/ extra space)
                (f"{footer_prefix}: #555 #444", ["#444", "#555"]),  # Space separated
                (
                    f"{footer_prefix}: #555;#444",
                    ["#444", "#555"],
                ),  # semicolon separated (w/o space)
                (
                    f"{footer_prefix}: #555; #444",
                    ["#444", "#555"],
                ),  # semicolon separated (w/ space)
                (
                    f"{footer_prefix}: #555 ; #444",
                    ["#444", "#555"],
                ),  # semicolon separated (w/ extra space)
                (
                    f"{footer_prefix}: #555/#444",
                    ["#444", "#555"],
                ),  # slash separated (w/o space)
                (
                    f"{footer_prefix}: #555/ #444",
                    ["#444", "#555"],
                ),  # slash separated (w/ space)
                (
                    f"{footer_prefix}: #555 / #444",
                    ["#444", "#555"],
                ),  # slash separated (w/ extra space)
                (
                    f"{footer_prefix}: #555&#444",
                    ["#444", "#555"],
                ),  # ampersand separated (w/o space)
                (
                    f"{footer_prefix}: #555& #444",
                    ["#444", "#555"],
                ),  # ampersand separated (w/ space)
                (
                    f"{footer_prefix}: #555 & #444",
                    ["#444", "#555"],
                ),  # ampersand separated (w/ extra space)
                (f"{footer_prefix}: #555 and #444", ["#444", "#555"]),  # and separated
                (
                    f"{footer_prefix}: #555, #444, and #333",
                    ["#333", "#444", "#555"],
                ),  # and separated
                # Mulitple issues (variant 2: multiple footers, supported by GitHub)
                (f"{footer_prefix}: #555\n{footer_prefix}: #444", ["#444", "#555"]),
                # More than 2 issues
                (
                    f"{footer_prefix}: #555, #444, #333",
                    ["#333", "#444", "#555"],
                ),
                # More than 2 issues (force numerical sort)
                (
                    f"{footer_prefix}: #555, #3333, #444",
                    ["#444", "#555", "#3333"],
                ),
                # Single issue listed multiple times
                (f"{footer_prefix}: #555, #555", ["#555"]),
                # Multiple footers with the same issue
                (f"{footer_prefix}: #555\n{footer_prefix}: #555", ["#555"]),
                # Multiple issues via multiple inline git footers
                (f"{footer_prefix}: #555, {footer_prefix}: #444", ["#444", "#555"]),
                # Multiple valid footers
                (
                    str.join(
                        "\n",
                        [
                            f"{footer_prefix}: #555",
                            "Signed-off-by: johndoe <johndoe@mail.com>",
                            f"{footer_prefix}: #444",
                        ],
                    ),
                    ["#444", "#555"],
                ),
                # ----------------------------------------- Invalid Sets ----------------------------------------- #
                # Must have colon because it is a git footer, these will not return a linked issue
                (f"{footer_prefix} #666", []),
                (f"{footer_prefix} #666, #777", []),
                # Invalid Multiple issues (although it is supported by GitHub, it is not supported by the parser)
                (f"{footer_prefix} #666, {footer_prefix} #777", []),
                # Invalid 'and' separation
                (f"{footer_prefix}: #666and#777", ["#666and#777"]),
                # Invalid prefix
                ("ref: #666", []),
                # body mentions an issue and has a different git footer
                (
                    "In #666, the devils in the details...\n\nSigned-off-by: johndoe <johndoe@mail.com>",
                    [],
                ),
            ]
        ],
        *[
            # JIRA style
            (
                f"feat(parser): add magic parser\n\n{footer}",
                linked_issues,
            )
            for footer_prefix in SUPPORTED_ISSUE_CLOSURE_PREFIXES
            for footer, linked_issues in [
                # Single issue
                (
                    f"{footer_prefix.capitalize()}: ABC-555",
                    ["ABC-555"],
                ),  # Git Footer style (capitalized)
                (f"{footer_prefix.lower()}: ABC-555", ["ABC-555"]),  # lowercase prefix
                (f"{footer_prefix.upper()}: ABC-555", ["ABC-555"]),  # uppercase prefix
                # Mulitple issues (variant 1: list with one prefix, not supported by GitHub)
                (
                    f"{footer_prefix}: ABC-555,ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # Comma separated (w/o space)
                (
                    f"{footer_prefix}: ABC-555, ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # Comma separated (w/ space)
                (
                    f"{footer_prefix}: ABC-555 , ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # Comma separated (w/ extra space)
                (
                    f"{footer_prefix}: ABC-555 ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # Space separated
                (
                    f"{footer_prefix}: ABC-555;ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # semicolon separated (w/o space)
                (
                    f"{footer_prefix}: ABC-555; ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # semicolon separated (w/ space)
                (
                    f"{footer_prefix}: ABC-555 ; ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # semicolon separated (w/ extra space)
                (
                    f"{footer_prefix}: ABC-555/ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # slash separated (w/o space)
                (
                    f"{footer_prefix}: ABC-555/ ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # slash separated (w/ space)
                (
                    f"{footer_prefix}: ABC-555 / ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # slash separated (w/ extra space)
                (
                    f"{footer_prefix}: ABC-555&ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # ampersand separated (w/o space)
                (
                    f"{footer_prefix}: ABC-555& ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # ampersand separated (w/ space)
                (
                    f"{footer_prefix}: ABC-555 & ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # ampersand separated (w/ extra space)
                (
                    f"{footer_prefix}: ABC-555 and ABC-444",
                    ["ABC-444", "ABC-555"],
                ),  # and separated
                (
                    f"{footer_prefix}: ABC-555, ABC-444, and ABC-333",
                    ["ABC-333", "ABC-444", "ABC-555"],
                ),  # and separated
                # Mulitple issues (variant 2: multiple footers, supported by GitHub)
                (
                    f"{footer_prefix}: ABC-555\n{footer_prefix}: ABC-444",
                    ["ABC-444", "ABC-555"],
                ),
                # More than 2 issues
                (
                    f"{footer_prefix}: ABC-555, ABC-444, ABC-333",
                    ["ABC-333", "ABC-444", "ABC-555"],
                ),
                # More than 2 issues (force numerical sort)
                (
                    f"{footer_prefix}: ABC-555, ABC-3333, ABC-444",
                    ["ABC-444", "ABC-555", "ABC-3333"],
                ),
                # Single issue listed multiple times
                (f"{footer_prefix}: ABC-555, ABC-555", ["ABC-555"]),
                # Multiple footers with the same issue
                (f"{footer_prefix}: ABC-555\n{footer_prefix}: ABC-555", ["ABC-555"]),
                # Multiple issues via multiple inline git footers
                (
                    f"{footer_prefix}: ABC-666, {footer_prefix}: ABC-777",
                    ["ABC-666", "ABC-777"],
                ),
                # Multiple valid footers
                (
                    str.join(
                        "\n",
                        [
                            f"{footer_prefix}: ABC-555",
                            "Signed-off-by: johndoe <johndoe@mail.com>",
                            f"{footer_prefix}: ABC-444",
                        ],
                    ),
                    ["ABC-444", "ABC-555"],
                ),
                # ----------------------------------------- Invalid Sets ----------------------------------------- #
                # Must have colon because it is a git footer, these will not return a linked issue
                (f"{footer_prefix} ABC-666", []),
                (f"{footer_prefix} ABC-666, ABC-777", []),
                # Invalid Multiple issues (although it is supported by GitHub, it is not supported by the parser)
                (f"{footer_prefix} ABC-666, {footer_prefix} ABC-777", []),
                # Invalid 'and' separation
                (f"{footer_prefix}: ABC-666andABC-777", ["ABC-666andABC-777"]),
                # Invalid prefix
                ("ref: ABC-666", []),
                # body mentions an issue and has a different git footer
                (
                    "In ABC-666, the devils in the details...\n\nSigned-off-by: johndoe <johndoe@mail.com>",
                    [],
                ),
            ]
        ],
        *[
            (
                f"feat(parser): add magic parser\n\n{footer}",
                linked_issues,
            )
            for footer, linked_issues in [
                # Multiple footers with the same issue but different prefixes
                ("Resolves: #555\nfix: #444", ["#444", "#555"]),
                # Whitespace padded footer
                ("  Resolves: #555\n", ["#555"]),
            ]
        ],
        (
            # Only grabs the issue reference when there is a GitHub PR reference in the subject
            "feat(parser): add magic parser (#123)\n\nCloses: #555",
            ["#555"],
        ),
        # Does not grab an issue when there is only a GitHub PR reference in the subject
        ("feat(parser): add magic parser (#123)", []),
        # Does not grab an issue when there is only a Bitbucket PR reference in the subject
        ("feat(parser): add magic parser (pull request #123)", []),
    ],
)
def test_parser_return_linked_issues_from_commit_message(
    default_conventional_parser: ConventionalCommitParser,
    message: str,
    linked_issues: Sequence[str],
    make_commit_obj: MakeCommitObjFn,
):
    parsed_results = default_conventional_parser.parse(make_commit_obj(message))

    assert isinstance(parsed_results, Iterable)
    assert 1 == len(parsed_results)

    result = next(iter(parsed_results))
    assert isinstance(result, ParsedCommit)
    assert tuple(linked_issues) == result.linked_issues


@pytest.mark.parametrize(
    "message, notices",
    [
        pytest.param(
            message,
            notices,
            id=test_id,
        )
        for test_id, message, notices in [
            (
                "single notice",
                dedent(
                    """\
                    fix(parser): fix regex in conventional parser

                    NOTICE: This is a notice
                    """
                ),
                ["This is a notice"],
            ),
            (
                "multiline notice",
                dedent(
                    """\
                    fix(parser): fix regex in conventional parser

                    NOTICE: This is a notice that is longer than
                    other notices
                    """
                ),
                ["This is a notice that is longer than other notices"],
            ),
            (
                "multiple notices",
                dedent(
                    """\
                    fix(parser): fix regex in conventional parser

                    NOTICE: This is a notice

                    NOTICE: This is a second notice
                    """
                ),
                ["This is a notice", "This is a second notice"],
            ),
            (
                "notice with other footer",
                dedent(
                    """\
                    fix(parser): fix regex in conventional parser

                    BREAKING CHANGE: This is a breaking change

                    NOTICE: This is a notice
                    """
                ),
                ["This is a notice"],
            ),
        ]
    ],
)
def test_parser_return_release_notices_from_commit_message(
    default_conventional_parser: ConventionalCommitParser,
    message: str,
    notices: Sequence[str],
    make_commit_obj: MakeCommitObjFn,
):
    parsed_results = default_conventional_parser.parse(make_commit_obj(message))

    assert isinstance(parsed_results, Iterable)
    assert len(parsed_results) == 1

    result = next(iter(parsed_results))
    assert isinstance(result, ParsedCommit)
    assert tuple(notices) == result.release_notices

    full_description = str.join("\n\n", result.descriptions)
    full_notice = str.join("\n\n", result.release_notices)
    assert full_notice not in full_description


##############################
# test custom parser options #
##############################
def test_parser_custom_default_level(make_commit_obj: MakeCommitObjFn):
    options = ConventionalCommitParserOptions(default_bump_level=LevelBump.MINOR)
    parsed_results = ConventionalCommitParser(options).parse(
        make_commit_obj("test(parser): add a test for conventional parser")
    )

    assert isinstance(parsed_results, Iterable)

    result = next(iter(parsed_results))
    assert isinstance(result, ParsedCommit)
    assert result.bump is LevelBump.MINOR


def test_parser_custom_allowed_types(
    default_conventional_parser: ConventionalCommitParser,
    make_commit_obj: MakeCommitObjFn,
):
    new_tag = "custom"
    custom_allowed_tags = [*default_conventional_parser.options.allowed_tags, new_tag]
    parser = ConventionalCommitParser(
        options=ConventionalCommitParserOptions(
            allowed_tags=tuple(custom_allowed_tags),
        )
    )

    for commit_type, commit_msg in [
        (new_tag, f"{new_tag}: ..."),  # no scope
        (new_tag, f"{new_tag}(parser): ..."),  # with scope
        ("chores", "chore(parser): ..."),  # existing, non-release tag
    ]:
        parsed_results = parser.parse(make_commit_obj(commit_msg))
        assert isinstance(parsed_results, Iterable)

        result = next(iter(parsed_results))
        assert isinstance(result, ParsedCommit)
        assert result.type == commit_type
        assert result.bump is LevelBump.NO_RELEASE


def test_parser_custom_allowed_types_ignores_non_types(
    default_conventional_parser: ConventionalCommitParser,
    make_commit_obj: MakeCommitObjFn,
):
    banned_tag = "feat"
    custom_allowed_tags = [*default_conventional_parser.options.allowed_tags]
    custom_allowed_tags.remove(banned_tag)

    parser = ConventionalCommitParser(
        options=ConventionalCommitParserOptions(
            allowed_tags=tuple(custom_allowed_tags),
        )
    )

    parsed_results = parser.parse(make_commit_obj(f"{banned_tag}(parser): ..."))
    assert isinstance(parsed_results, Iterable)

    result = next(iter(parsed_results))
    assert isinstance(result, ParseError)


def test_parser_custom_minor_tags(make_commit_obj: MakeCommitObjFn):
    custom_minor_tag = "docs"
    parser = ConventionalCommitParser(
        options=ConventionalCommitParserOptions(minor_tags=(custom_minor_tag,))
    )

    parsed_results = parser.parse(make_commit_obj(f"{custom_minor_tag}: ..."))
    assert isinstance(parsed_results, Iterable)

    result = next(iter(parsed_results))
    assert isinstance(result, ParsedCommit)
    assert result.bump is LevelBump.MINOR


def test_parser_custom_patch_tags(make_commit_obj: MakeCommitObjFn):
    custom_patch_tag = "test"
    parser = ConventionalCommitParser(
        options=ConventionalCommitParserOptions(patch_tags=(custom_patch_tag,))
    )

    parsed_results = parser.parse(make_commit_obj(f"{custom_patch_tag}: ..."))
    assert isinstance(parsed_results, Iterable)

    result = next(iter(parsed_results))
    assert isinstance(result, ParsedCommit)
    assert result.bump is LevelBump.PATCH


def test_parser_ignore_merge_commit(
    default_conventional_parser: ConventionalCommitParser,
    make_commit_obj: MakeCommitObjFn,
):
    # Setup: Enable parsing of linked issues
    parser = ConventionalCommitParser(
        options=ConventionalCommitParserOptions(
            **{
                **default_conventional_parser.options.__dict__,
                "ignore_merge_commits": True,
            }
        )
    )

    base_commit = make_commit_obj("Merge branch 'fix/fix-feature' into 'main'")
    incomming_commit = make_commit_obj("feat: add a new feature")

    # Setup: Create a merge commit
    merge_commit = make_commit_obj("Merge branch 'feat/add-new-feature' into 'main'")
    merge_commit.parents = [base_commit, incomming_commit]

    # Action
    parsed_result = parser.parse(merge_commit)

    assert isinstance(parsed_result, ParseError)
    assert "Ignoring merge commit" in parsed_result.error
